"""Tests for fetching surfaces."""
from geomeppy.idf import EpBunch


class TestSurfaces:
    def test_surfaces(self, wwr_idf):
        # type: () -> None
        idf = wwr_idf
        surfaces = idf.getsurfaces()
        assert surfaces
        assert all(isinstance(s, EpBunch) for s in surfaces)

    def test_subsurfaces(self, wwr_idf):
        # type: () -> None
        idf = wwr_idf
        surfaces = idf.getsubsurfaces()
        assert surfaces
        assert all(isinstance(s, EpBunch) for s in surfaces)

    def test_shadingsurfaces(self, base_idf):
        # type: () -> None
        idf = base_idf
        surfaces = idf.getshadingsurfaces()
        assert surfaces
        assert all(isinstance(s, EpBunch) for s in surfaces)



import pytest

from geomeppy.geom.surfaces import get_adjacencies
from geomeppy.geom.surfaces import set_coords


@pytest.fixture
def shadow_intersecting():
    shadow_blocks = [
        {
            "name": "PN1001_Bld1004_EL24",
            "coordinates": [
                (-83635.2890999997, -100958.99369999953),
                (-83636.50970000029, -100976.35019999929),
                (-83651.5382000003, -100975.29350000061),
                (-83650.31759999972, -100957.93679999933),
                (-83635.2890999997, -100958.99369999953),
            ],
            "height": 21.0,
        }
    ]
    zones = [
        {
            "name": "PN1001_Bld1003 Zone3",
            "coordinates": [
                (-83614.40199999977, -100952.4587999992),
                (-83616.24896021019, -100960.96199999936),
                (-83635.42752116646, -100960.96199999936),
                (-83635.2890999997, -100958.99369999953),
                (-83634.0064000003, -100940.75280000083),
                (-83614.40199999977, -100952.4587999992),
            ],
            "height": 3.0,
            "num_stories": 1,
        },
        {
            "name": "PN1001_Bld1003 Zone4",
            "coordinates": [
                (-83616.24896021019, -100960.96199999936),
                (-83619.72295787116, -100976.95590000041),
                (-83636.55229433342, -100976.95590000041),
                (-83636.50970000029, -100976.35019999929),
                (-83635.42752116646, -100960.96199999936),
                (-83616.24896021019, -100960.96199999936),
            ],
            "height": 3.0,
            "num_stories": 1,
        },
    ]
    return {"zones": zones, "shadows": shadow_blocks}


def _test_adjacencies(new_idf, shadow_intersecting):
    """Test in a single plane, but angled."""
    try:
        ggr = new_idf.idfobjects["GLOBALGEOMETRYRULES"][0]
    except IndexError:
        ggr = None
    for block in shadow_intersecting["shadows"]:
        new_idf.add_shading_block(**block)
    for block in shadow_intersecting["zones"]:
        new_idf.add_block(**block)
    new_idf.translate_to_origin()
    # new_idf.view_model()
    adjacencies = get_adjacencies(new_idf.getsurfaces() + new_idf.getshadingsurfaces())
    done = []
    for surface in adjacencies:
        key, name = surface
        new_surfaces = adjacencies[surface]
        old_obj = new_idf.getobject(key.upper(), name)
        for i, new_coords in enumerate(new_surfaces, 1):
            if (key, name, new_coords) in done:
                continue
            done.append((key, name, new_coords))
            new = new_idf.copyidfobject(old_obj)
            new.Name = "%s_%i" % (name, i)
            set_coords(new, new_coords, ggr)
        new_idf.removeidfobject(old_obj)
    assert len(new_idf.getshadingsurfaces()) == 6
    assert len(new_idf.getsurfaces("wall")) == 10


def test_two_walls_one_shadow(new_idf):
    wall1 = [(0, 0.1, 0), (0, 1, 0), (0, 1, 1), (0, 0.1, 1)]
    wall2 = [(0, 1, 0), (0, 2.2, 0), (0, 2.2, 1), (0, 1, 1)]
    shadow = reversed([(0, 0, 0), (0, 2, 0), (0, 2, 2), (0, 0, 2)])
    w1 = new_idf.newidfobject(
        "BUILDINGSURFACE:DETAILED", Name="w1", Surface_Type="wall"
    )
    w1.setcoords(wall1)
    w2 = new_idf.newidfobject(
        "BUILDINGSURFACE:DETAILED", Name="w2", Surface_Type="wall"
    )
    w2.setcoords(wall2)
    sh = new_idf.newidfobject("SHADING:ZONE:DETAILED", Name="sh")
    sh.setcoords(shadow)
    new_idf.intersect()
    new_idf.match()
    adiabatic = [
        w
        for w in new_idf.getsurfaces("wall")
        if w.Outside_Boundary_Condition == "adiabatic"
    ]
    assert len(adiabatic) == 2
    assert len(new_idf.getsurfaces("wall")) == 3
    assert len(new_idf.getshadingsurfaces()) == 3



from geomeppy.extractor import copy_constructions, copy_geometry


def test_copy_geometry(extracts_idf):
    idf = copy_geometry(extracts_idf)
    assert idf.getobject("ZONE", "z1 Thermal Zone")
    assert not idf.getobject("MATERIAL", "Spam")


def test_copy_constructions(extracts_idf):
    idf = copy_constructions(extracts_idf)
    assert not idf.getobject("ZONE", "z1 Thermal Zone")
    assert idf.getobject("MATERIAL", "Spam")



"""Core and perimeter zoning approach."""
from itertools import product
from geomeppy.geom.polygons import Polygon2D


def get_core(footprint, perim_depth=None):
    poly = Polygon2D(footprint)
    core = poly.buffer(distance=-perim_depth)
    return core


def get_perims(footprint, core):
    perims = []
    poly = Polygon2D(footprint)
    for edge in poly.edges:
        c1 = sorted(
            product([edge.p1] * len(core), core),
            key=lambda x: x[0].relative_distance(x[1]),
        )[0][1]
        c2 = sorted(
            product([edge.p2] * len(core), core),
            key=lambda x: x[0].relative_distance(x[1]),
        )[0][1]
        perims.append(Polygon2D([c1, edge.p1, edge.p2, c2]))
    return perims


def core_perim_zone_coordinates(footprint, perim_depth):
    """Returns a dictionary with the coordinates of the core/perimeter zones and
    a list of tuples containing the coordinates for the core zone."""
    zones_perim = get_perims(footprint, get_core(footprint, perim_depth))
    core = get_core(footprint, perim_depth)

    zones_dict = {}
    for idx, zone_perim in enumerate(zones_perim, 1):
        zone_name = "Perimeter_Zone_%i" % idx
        perim_zones_coords = []
        for pts in zone_perim:
            perim_zones_coords.append(pts.as_tuple(dims=2))
        zones_dict[zone_name] = perim_zones_coords
    core_zone_coords = []
    for pts in core:
        core_zone_coords.append(pts.as_tuple(dims=2))

    zones_dict["Core_Zone"] = core_zone_coords
    return zones_dict, core_zone_coords



"""A module containing pytest fixtures that are used in multiple places in the test suite."""
import os

import matplotlib
import pytest
from eppy.iddcurrent import iddcurrent
from six import StringIO

from geomeppy.idf import IDF

if not os.getenv("CI"):
    matplotlib.use("Qt5Agg")

base_idf_txt = """
    Version, 8.5;
    Building, Building 1, , , , , , , ;
    Zone, z1 Thermal Zone, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
    Zone, z2 Thermal Zone, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
    BuildingSurface:Detailed, z1_FLOOR, Floor, , z1 Thermal Zone, ground, , NoSun, NoWind, , , 1.0, 2.1, 0.0, 2.0, 2.0, 0.0, 2.0, 1.0, 0.0, 1.0, 1.1, 0.0;
    BuildingSurface:Detailed, z1_ROOF, Roof, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 1.0, 0.5, 2.0, 2.0, 0.5, 1.0, 2.1, 0.5, 1.0, 1.1, 0.5;
    BuildingSurface:Detailed, z1_WALL_0001, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.0, 1.1, 0.5, 1.0, 1.1, 0.0, 1.0, 2.1, 0.0, 1.0, 2.1, 0.5;
    BuildingSurface:Detailed, z1_WALL_0002, Wall, , z1 Thermal Zone, Outdoors, , SunExposed, WindExposed, , , 1.0, 2.1, 0.5, 1.0, 2.1, 0.0, 2.0, 2.0, 0.0, 2.0, 2.0, 0.5;
    BuildingSurface:Detailed, z1_WALL_0003, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 2.0, 0.5, 2.0, 2.0, 0.0, 2.0, 1.0, 0.0, 2.0, 1.0, 0.5;
    BuildingSurface:Detailed, z1_WALL_0004, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 1.0, 0.5, 2.0, 1.0, 0.0, 1.0, 1.1, 0.0, 1.0, 1.1, 0.5;
    BuildingSurface:Detailed, z2_FLOOR, Floor, , z2 Thermal Zone, ground, , NoSun, NoWind, , , 1.5, 3.05, 0.0, 2.5, 2.95, 0.0, 2.5, 1.95, 0.0, 1.5, 2.05, 0.0;
    BuildingSurface:Detailed, z2_ROOF, Roof, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.5, 1.95, 0.5, 2.5, 2.95, 0.5, 1.5, 3.05, 0.5, 1.5, 2.05, 0.5;
    BuildingSurface:Detailed, z2_WALL_0001, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.5, 2.05, 0.5, 1.5, 2.05, 0.0, 1.5, 3.05, 0.0, 1.5, 3.05, 0.5;
    BuildingSurface:Detailed, z2_WALL_0002, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.5, 3.05, 0.5, 1.5, 3.05, 0.0, 2.5, 2.95, 0.0, 2.5, 2.95, 0.5;
    BuildingSurface:Detailed, z2_WALL_0003, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.5, 2.95, 0.5, 2.5, 2.95, 0.0, 2.5, 1.95, 0.0, 2.5, 1.95, 0.5;
    BuildingSurface:Detailed, z2_WALL_0004, Wall, , z2 Thermal Zone, Outdoors, , SunExposed, WindExposed, , , 2.5, 1.95, 0.5, 2.5, 1.95, 0.0, 1.5, 2.05, 0.0, 1.5, 2.05, 0.5;
    Shading:Zone:Detailed, z2_SHADE_0003, z2_WALL_0003, , 4, 2.5, 2.95, 0.5, 2.6, 2.95, 0.3, 2.6, 1.95, 0.3, 2.5, 1.95, 0.5;
    """


@pytest.fixture()
def base_idf():
    """A two-zone model with a shading surface on one wall."""
    iddfhandle = StringIO(iddcurrent.iddtxt)
    if IDF.getiddname() == None:
        IDF.setiddname(iddfhandle)

    return IDF(StringIO(base_idf_txt))


ring_idf_txt = """
    Version, 8.5;
    Building, Building 1, , , , , , , ;
    Zone, Thermal Zone 1, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
    Zone, Thermal Zone 2, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
    BuildingSurface:Detailed, z1 Floor 0001, Floor, , Thermal Zone 2, Ground, , NoSun, NoWind, , , 0.0, 2.9, 0.0, 0.0, 0.0, 0.0, -2.14, 0.0, 0.0, -2.14, 2.9, 0.0;
    BuildingSurface:Detailed, z1 Wall 0001, Wall, , Thermal Zone 2, Outdoors, , SunExposed, WindExposed, , , -2.14, 0.0, 1.5, -2.14, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.5;
    BuildingSurface:Detailed, z1 Wall 0002, Wall, , Thermal Zone 2, Outdoors, , SunExposed, WindExposed, , , -2.14, 2.9, 1.5, -2.14, 2.9, 0.0, -2.14, 0.0, 0.0, -2.14, 0.0, 1.5;
    BuildingSurface:Detailed, z1 Wall 0003, Wall, , Thermal Zone 2, Outdoors, , SunExposed, WindExposed, , , 0.0, 2.9, 1.5, 0.0, 2.9, 0.0, -2.14, 2.9, 0.0, -2.14, 2.9, 1.5;
    BuildingSurface:Detailed, z1 Wall 0004, Wall, , Thermal Zone 2, Outdoors, , SunExposed, WindExposed, , , 0.0, 0.0, 1.5, 0.0, 0.0, 0.0, 0.0, 2.9, 0.0, 0.0, 2.9, 1.5;
    BuildingSurface:Detailed, z1 Roof 0001, Roof, , Thermal Zone 2, Outdoors, , SunExposed, WindExposed, , , 0.0, 0.0, 1.5, 0.0, 2.9, 1.5, -2.14, 2.9, 1.5, -2.14, 0.0, 1.5;
    
    BuildingSurface:Detailed, z2 Floor 0001, Floor, , Thermal Zone 1, Ground, , NoSun, NoWind, , , -0.259, 2.46, 1.5, -0.259, 0.4, 1.5, -1.68, 0.4, 1.5, -1.68, 2.46, 1.5;
    BuildingSurface:Detailed, z2 Wall 0001, Wall, , Thermal Zone 1, Outdoors, , SunExposed, WindExposed, , , -0.259, 2.46, 3.0, -0.259, 2.46, 1.5, -1.68, 2.46, 1.5, -1.68, 2.46, 3.0;
    BuildingSurface:Detailed, z2 Wall 0002, Wall, , Thermal Zone 1, Outdoors, , SunExposed, WindExposed, , , -0.259, 0.4, 3.0, -0.259, 0.4, 1.5, -0.259, 2.46, 1.5, -0.259, 2.46, 3.0;
    BuildingSurface:Detailed, z2 Wall 0003, Wall, , Thermal Zone 1, Outdoors, , SunExposed, WindExposed, , , -1.68, 0.4, 3.0, -1.68, 0.4, 1.5, -0.259, 0.4, 1.5, -0.259, 0.4, 3.0;
    BuildingSurface:Detailed, z2 Wall 0004, Wall, , Thermal Zone 1, Outdoors, , SunExposed, WindExposed, , , -1.68, 2.46, 3.0, -1.68, 2.46, 1.5, -1.68, 0.4, 1.5, -1.68, 0.4, 3.0;
    BuildingSurface:Detailed, z2 Roof 0001, Roof, , Thermal Zone 1, Outdoors, , SunExposed, WindExposed, , , -0.259, 0.4, 3.0, -0.259, 2.46, 3.0, -1.68, 2.46, 3.0, -1.68, 0.4, 3.0;
    """


@pytest.fixture()
def ring_idf():
    """A two-zone model with an inner ring."""
    iddfhandle = StringIO(iddcurrent.iddtxt)
    if IDF.getiddname() == None:
        IDF.setiddname(iddfhandle)

    return IDF(StringIO(ring_idf_txt))


@pytest.fixture()
def new_idf():
    """An empty model."""
    iddfhandle = StringIO(iddcurrent.iddtxt)
    IDF.setiddname(iddfhandle, testing=True)
    idf = IDF()
    idf.new()
    return idf


@pytest.fixture()
def wwr_idf(new_idf):
    """A model with just two walls and a window."""
    test_walls = (
        [[0, 0, 0], [0, 1, 0], [0, 1, 1], [0, 0, 1]],
        [[0, 0, 0], [1, 0, 0], [1, 0, 1], [0, 0, 1]],
    )
    for i, coords in enumerate(test_walls, 1):
        new_idf.newidfobject(
            "FENESTRATIONSURFACE:DETAILED",
            Name="window%s" % i,
            Surface_Type="window",
            Construction_Name="ExtWindow",
            Building_Surface_Name="wall%s" % i,
        )
        wall = new_idf.newidfobject(
            "BUILDINGSURFACE:DETAILED",
            Name="wall%s" % i,
            Surface_Type="wall",
            Outside_Boundary_Condition="Outdoors",
        )
        wall.setcoords(coords)
    return new_idf


extracts_idf_txt = """
    Version, 8.8;
    Zone, z1 Thermal Zone, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
    Material, Spam, Rough, 0.1, 1, 1000, 1000, 0.9, 0.9, 0.9;
    """


@pytest.fixture()
def extracts_idf():
    """A model with a minimal zone and material object."""
    iddfhandle = StringIO(iddcurrent.iddtxt)
    if IDF.getiddname() == None:
        IDF.setiddname(iddfhandle)

    return IDF(StringIO(extracts_idf_txt))



"""Tests for Segment class, representing a line segment."""
from geomeppy.geom.segments import Segment
from geomeppy.geom.vectors import Vector3D


def test_segment_repr():
    # type: () -> None
    edge = Segment(Vector3D(0, 0, 0), Vector3D(2, 0, 0))
    assert eval(repr(edge)) == edge


def test_collinear():
    # type: () -> None
    # same line
    edge1 = Segment(Vector3D(0, 0, 0), Vector3D(1, 1, 1))
    edge2 = Segment(Vector3D(0, 0, 0), Vector3D(1, 1, 1))
    assert edge1._is_collinear(edge2)

    # opposite direction line
    edge1 = Segment(Vector3D(1, 1, 1), Vector3D(0, 0, 0))
    edge2 = Segment(Vector3D(0, 0, 0), Vector3D(1, 1, 1))
    assert edge1._is_collinear(edge2)

    # edge1 is longer
    edge1 = Segment(Vector3D(0, 0, 0), Vector3D(4, 4, 4))
    edge2 = Segment(Vector3D(1, 1, 1), Vector3D(2, 2, 2))
    assert edge1._is_collinear(edge2)

    # same start point, different lengths
    edge1 = Segment(Vector3D(0, 0, 0), Vector3D(1, 1, 1))
    edge2 = Segment(Vector3D(0, 0, 0), Vector3D(2, 2, 2))
    assert edge1._is_collinear(edge2)

    # something being missed
    edge1 = Segment(Vector3D(1, 4, 0), Vector3D(1, 0, 0))
    edge2 = Segment(Vector3D(1, 0, 0), Vector3D(1, 2, 0))
    assert edge1._is_collinear(edge2)

    # parallel
    edge1 = Segment(Vector3D(0, 0, 0), Vector3D(1, 1, 1))
    edge2 = Segment(Vector3D(1, 0, 0), Vector3D(2, 1, 1))
    assert not edge1._is_collinear(edge2)



"""Tests for recipes."""
from geomeppy.idf import IDF
from geomeppy.geom.intersect_match import intersect_idf_surfaces, match_idf_surfaces
from geomeppy.geom.polygons import Polygon3D
from geomeppy.geom.vectors import Vector2D, Vector3D
from geomeppy.recipes import rotate, set_wwr, translate, translate_to_origin
from geomeppy.utilities import almostequal
from geomeppy.view_geometry import _get_collections, _get_shading, _get_surfaces


class TestTranslate:
    def test_translate(self, base_idf):
        # type: (IDF) -> None
        idf = base_idf
        surfaces = idf.getsurfaces()
        translate(surfaces, (50, 100))  # move to x + 50, y + 100
        result = Polygon3D(surfaces[0].coords).xs
        expected = [52.0, 52.0, 51.0, 51.0]
        assert result == expected
        translate(surfaces, (-50, -100))  # move back

        translate(surfaces, Vector2D(50, 100))  # move to x + 50, y + 100
        result = Polygon3D(surfaces[0].coords).xs
        expected = [52.0, 52.0, 51.0, 51.0]
        assert result == expected
        translate(surfaces, (-50, -100))  # move back

        translate(surfaces, Vector3D(50, 100, 0))  # move to x + 50, y + 100
        result = Polygon3D(surfaces[0].coords).xs
        expected = [52.0, 52.0, 51.0, 51.0]
        assert result == expected

    def test_translate_to_origin(self, base_idf):
        # type: (IDF) -> None
        idf = base_idf
        surfaces = idf.getsurfaces()
        translate(surfaces, (50000, 10000))  # move to x + 50, y + 100
        result = Polygon3D(surfaces[0].coords).xs
        expected = [50002.0, 50002.0, 50001.0, 50001.0]
        assert result == expected

        translate_to_origin(idf)
        surfaces = idf.getsurfaces()
        min_x = min(min(Polygon3D(s.coords).xs) for s in surfaces)
        min_y = min(min(Polygon3D(s.coords).ys) for s in surfaces)
        assert min_x == 0
        assert min_y == 0

    def test_rotate_360(self, base_idf):
        # type: (IDF) -> None
        idf = base_idf
        surface = idf.getsurfaces()[0]
        coords = [Vector3D(*v) for v in [(0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0)]]
        surface.setcoords(coords)
        expected = surface.coords
        rotate([surface], 360)
        result = surface.coords
        assert almostequal(result, expected)

    def test_rotate_idf_360(self, base_idf):
        # type: (IDF) -> None
        idf1 = base_idf
        idf2 = IDF()
        idf2.initreadtxt(idf1.idfstr())
        idf1.rotate(360)
        floor1 = Polygon3D(idf1.getsurfaces("floor")[0].coords).normalize_coords(None)
        floor2 = Polygon3D(idf2.getsurfaces("floor")[0].coords).normalize_coords(None)
        assert almostequal(floor1, floor2)
        shade1 = Polygon3D(idf1.getshadingsurfaces()[0].coords).normalize_coords(None)
        shade2 = Polygon3D(idf1.getshadingsurfaces()[0].coords).normalize_coords(None)
        assert almostequal(shade1, shade2)

    def test_centre(self, base_idf):
        # type: (IDF) -> None
        idf = base_idf
        result = idf.centroid
        expected = Vector2D(1.75, 2.025)
        assert result == expected

    def test_scale_idf(self, base_idf):
        # type: (IDF) -> None
        idf1 = base_idf
        idf2 = IDF()
        idf2.initreadtxt(idf1.idfstr())
        idf1.scale(10)
        idf1.scale(0.1)
        floor1 = Polygon3D(idf1.getsurfaces("floor")[0].coords).normalize_coords(None)
        floor2 = Polygon3D(idf2.getsurfaces("floor")[0].coords).normalize_coords(None)
        assert almostequal(floor1, floor2)


class TestMatchSurfaces:
    def test_set_wwr(self, base_idf):
        # type: (IDF) -> None
        idf = base_idf
        intersect_idf_surfaces(idf)
        match_idf_surfaces(idf)
        wwr = 0.25
        set_wwr(idf, wwr)
        windows = idf.idfobjects["FENESTRATIONSURFACE:DETAILED"]
        assert len(windows) == 8
        for window in windows:
            wall = idf.getobject(
                "BUILDINGSURFACE:DETAILED", window.Building_Surface_Name
            )
            assert almostequal(window.area, wall.area * wwr, 3)


class TestViewGeometry:
    def test_get_surfaces(self, base_idf):
        # type: (IDF) -> None
        idf = base_idf
        surfaces = _get_surfaces(idf)
        assert len(surfaces) == 12

    def test_get_shading(self, base_idf):
        # type: (IDF) -> None
        idf = base_idf
        shading = _get_shading(idf)
        assert len(shading) == 1

    def test_get_collections(self, base_idf):
        # type: (IDF) -> None
        idf = base_idf
        collections = _get_collections(idf)
        assert len(collections) == 5
        for c in collections:
            assert c

    def test_view_model(self, base_idf):
        # type: (IDF) -> None
        idf = base_idf
        idf.view_model(test=True)


class TestWWR:
    def is_expected_wwr(self, idf, wwr):
        windows_area = sum(w.area for w in idf.getsubsurfaces("window"))
        walls_area = sum(w.area for w in idf.getsurfaces("wall"))
        expected_area = walls_area * wwr
        return almostequal(windows_area, expected_area, 3)

    def test_wwr(self, wwr_idf):
        idf = wwr_idf
        wwr = 0.2
        idf.set_wwr(wwr)
        assert self.is_expected_wwr(idf, wwr), "Not all walls have the expected WWR set"

    def test_wwr_map(self, wwr_idf):
        idf = wwr_idf
        expected_wwr = 0.2
        idf.set_wwr(wwr=0.99, wwr_map={90.0: 0.1, 180: 0.3})
        assert self.is_expected_wwr(
            idf, expected_wwr
        ), "Not all walls have the expected WWR set"

    def test_wwr_zero(self, wwr_idf):
        idf = wwr_idf
        idf.set_wwr(wwr=0, wwr_map={90.0: 0.1})
        expected_wwr = 0.05
        assert self.is_expected_wwr(
            idf, expected_wwr
        ), "Not all walls have the expected WWR set"

    def test_wwr_none(self, wwr_idf):
        idf = wwr_idf
        idf.set_wwr(wwr=None, wwr_map={90.0: 0.3})
        expected_wwr = 0.15
        assert self.is_expected_wwr(
            idf, expected_wwr
        ), "Not all walls have the expected WWR set"

    def test_wwr_two_window_constructions(self, wwr_idf):
        idf = wwr_idf
        idf.newidfobject(
            "FENESTRATIONSURFACE:DETAILED",
            Name="window2",
            Surface_Type="window",
            Construction_Name="ExtWindow2",
            Building_Surface_Name="wall1",
        )
        wwr = 0.2
        try:
            idf.set_wwr(wwr)
            assert (
                False
            ), "Should have raised an error since windows with more than one construction are present"
        except ValueError:
            pass
        idf.set_wwr(wwr, construction="ExtWindow")
        assert self.is_expected_wwr(idf, wwr)

    def test_wwr_mixed_subsurfaces(self, wwr_idf):
        idf = wwr_idf
        idf.newidfobject(
            "DOOR",
            Name="door1",
            Construction_Name="ExtDoor",
            Building_Surface_Name="wall1",
        )
        wwr = 0.2
        try:
            idf.set_wwr(wwr)
            assert (
                False
            ), "Should have raised an error since not all subsurfaces are windows"
        except ValueError:
            pass
        idf.set_wwr(wwr, force=True)
        assert self.is_expected_wwr(idf, wwr)

    def test_wwr_by_orientation(self, wwr_idf):
        idf = wwr_idf
        idf.set_wwr(wwr=0.2, orientation="south")  # there is a south-facing wall
        idf.set_wwr(wwr=0.2, orientation="west")  # there is no west-facing wall
        expected_wwr = 0.1
        assert self.is_expected_wwr(
            idf, expected_wwr
        ), "WWR not set correctly by orientation"



import os
import shutil

import pytest

from geomeppy import IDF


@pytest.fixture
def tmp_dir():
    os.mkdir("tests/tutorial")
    yield
    shutil.rmtree("tests/tutorial")


@pytest.mark.usefixtures("tmp_dir")
def test_tutorial():
    IDF.setiddname("C:/EnergyPlusV9-1-0/Energy+.idd", testing=True)
    idf = IDF("C:/EnergyPlusV9-1-0/ExampleFiles/Minimal.idf")
    idf.epw = "USA_CA_San.Francisco.Intl.AP.724940_TMY3.epw"
    idf.add_block(
        name="Boring hut", coordinates=[(10, 0), (10, 10), (0, 10), (0, 0)], height=3.5
    )
    idf.set_default_constructions()
    idf.intersect_match()
    idf.set_wwr(0.6, construction="Project External Window")
    idf.to_obj("tests/tutorial/boring_hut.obj")
    idf.run(output_directory="tests/tutorial")



"""
This module contains the implementation of `geomeppy.IDF`.
"""
import itertools
from typing import Any, Dict, List, Optional, Union  # noqa

from eppy.bunch_subclass import EpBunch  # noqa
from eppy.idf_msequence import Idf_MSequence  # noqa

from .geom.intersect_match import intersect_idf_surfaces, match_idf_surfaces
from .builder import Block, Zone
from .geom.polygons import bounding_box, Polygon2D  # noqa
from .geom.vectors import Vector2D, Vector3D  # noqa
from .io.obj import export_to_obj
from .patches import PatchedIDF
from .recipes import (
    set_default_constructions,
    set_wwr,
    rotate,
    scale,
    translate,
    translate_to_origin,
)
from .view_geometry import view_idf
from .geom.core_perim import core_perim_zone_coordinates


def new_idf(fname):
    # type: (str) -> IDF
    """Create a new blank IDF.

    :param fname: A name for the new IDF.

    """
    idf = IDF()
    idf.new(fname)
    return idf


class IDF(PatchedIDF):
    """Geometry-enabled IDF class, usable in the same way as Eppy's IDF.

    This adds geometry functionality to Eppy's IDF class.

    """

    def intersect_match(self):
        # type: () -> None
        """Intersect all surfaces in the IDF, then set boundary conditions."""
        self.intersect()
        self.match()

    def intersect(self):
        # type: () -> None
        """Intersect all surfaces in the IDF."""
        intersect_idf_surfaces(self)

    def match(self):
        # type: () -> None
        """Set boundary conditions for all surfaces in the IDF."""
        match_idf_surfaces(self)

    def translate_to_origin(self):
        # type: () -> None
        """Move an IDF close to the origin so that it can be viewed in SketchUp."""
        translate_to_origin(self)

    def translate(self, vector):
        # type: (Vector2D) -> None
        """Move the IDF in the direction given by a vector.

        :param vector: A vector to translate by.

        """
        surfaces = self.getsurfaces()
        translate(surfaces, vector)
        subsurfaces = self.getsubsurfaces()
        translate(subsurfaces, vector)
        shadingsurfaces = self.getshadingsurfaces()
        translate(shadingsurfaces, vector)

    def rotate(self, angle, anchor=None):
        # type: (float, Optional[Union[Vector2D, Vector3D]]) -> None
        """Rotate the IDF counterclockwise by the angle given.

        :param angle: Angle (in degrees) to rotate by.
        :param anchor: Point around which to rotate. Default is the centre of the the IDF's bounding box.

        """
        anchor = anchor or self.centroid
        surfaces = self.getsurfaces()
        subsurfaces = self.getsubsurfaces()
        shadingsurfaces = self.getshadingsurfaces()
        self.translate(-anchor)
        rotate(surfaces, angle)
        rotate(subsurfaces, angle)
        rotate(shadingsurfaces, angle)
        self.translate(anchor)

    def scale(self, factor, anchor=None, axes="xy"):
        # type: (float, Optional[Union[Vector2D, Vector3D]], str) -> None
        """Scale the IDF by a scaling factor.

        :param factor: Factor to scale by.
        :param anchor: Point to scale around. Default is the centre of the the IDF's bounding box.
        :param axes: Axes to scale on. Default 'xy'.

        """
        anchor = anchor or self.centroid
        surfaces = self.getsurfaces()
        subsurfaces = self.getsubsurfaces()
        shadingsurfaces = self.getshadingsurfaces()
        self.translate(-anchor)
        scale(surfaces, factor, axes)
        scale(subsurfaces, factor, axes)
        scale(shadingsurfaces, factor, axes)
        self.translate(anchor)

    def set_default_constructions(self):
        # type: () -> None
        set_default_constructions(self)

    def bounding_box(self):
        # type: () -> Polygon2D
        """Calculate the site bounding box.

        :returns: A polygon of the bounding box.

        """
        floors = self.getsurfaces("floor")
        return bounding_box(floors)

    @property
    def centroid(self):
        # type: () -> Vector2D
        """Calculate the centroid of the site bounding box.

        :returns: The centroid of the site bounding box.

        """
        bbox = self.bounding_box()
        return bbox.centroid

    def getsurfaces(self, surface_type=""):
        # type: (str) -> Union[List[EpBunch], Idf_MSequence]
        """Return all surfaces in the IDF.

        :param surface_type: Type of surface to get. Defaults to all.
        :returns: IDF surfaces.

        """
        surfaces = itertools.chain.from_iterable(
            [
                self.idfobjects[key.upper()]
                for key in self.idd_index["ref2names"]["SurfaceNames"]
            ]
        )
        if surface_type:
            surfaces = filter(
                lambda x: x.Surface_Type.lower() == surface_type.lower(), surfaces
            )
        return list(surfaces)

    def getsubsurfaces(self, surface_type=""):
        # type: (str) -> Union[List[EpBunch], Idf_MSequence]
        """Return all subsurfaces in the IDF.

        :param surface_type: Type of surface to get. Defaults to all.
        :returns: IDF surfaces.

        """
        surfaces = itertools.chain.from_iterable(
            [
                self.idfobjects[key.upper()]
                for key in self.idd_index["ref2names"]["SubSurfNames"]
            ]
        )
        if surface_type:
            surfaces = filter(
                lambda x: x.Surface_Type.lower() == surface_type.lower(), surfaces
            )
        return list(surfaces)

    def getshadingsurfaces(self, surface_type=""):
        # type: (str) -> Union[List[EpBunch], Idf_MSequence]
        """Return all subsurfaces in the IDF.

        :param surface_type: Type of surface to get. Defaults to all.
        :returns: IDF surfaces.

        """
        surfaces = itertools.chain.from_iterable(
            [
                self.idfobjects[key.upper()]
                for key in self.idd_index["ref2names"]["AllShadingSurfNames"]
            ]
        )
        if surface_type:
            surfaces = filter(
                lambda x: x.Surface_Type.lower() == surface_type.lower(), surfaces
            )
        return list(surfaces)

    def set_wwr(
        self, wwr=0.2, construction=None, force=False, wwr_map={}, orientation=None
    ):
        # type: (Optional[float], Optional[str], Optional[bool], Optional[dict], Optional[str]) -> None
        """Add strip windows to all external walls.

        Different WWR can be applied to specific wall orientations using the `wwr_map` keyword arg.
        This map is a dict of wwr values, keyed by `wall.azimuth`, which overrides the default passed as `wwr`.

        They can also be applied to walls oriented to a compass point, e.g. north, which will apply to walls which
        have an azimuth within 45 degrees of due north.

        :param wwr: Window to wall ratio in the range 0.0 to 1.0.
        :param construction: Name of a window construction.
        :param force: True to remove all subsurfaces before setting the WWR.
        :param wwr_map: Mapping from wall orientation (azimuth) to WWR, e.g. {180: 0.25, 90: 0.2}.
        :param orientation: One of "north", "east", "south", "west". Walls within 45 degrees will be affected.

        """
        set_wwr(self, wwr, construction, force, wwr_map, orientation)

    def view_model(self, test=False):
        # type: (Optional[bool]) -> None
        """Show a zoomable, rotatable representation of the IDF."""
        view_idf(idf_txt=self.idfstr(), test=test)

    def to_obj(self, fname=None, mtllib=None):
        # type: (Optional[str], Optional[str]) -> None
        """Export an OBJ file representation of the IDF.

        This can be used for viewing in tools which support the .obj format.

        :param fname: A filename for the .obj file. If None we try to base it on IDF.idfname and change the filetype.
        :param mtllib: The name of a .mtl file to be referenced from the .obj file. If None, we use default.mtl.
        """
        if not fname:
            try:
                fname = self.idfname.replace(".idf", ".obj")
            except AttributeError:
                fname = "default.obj"
        export_to_obj(self, fname, mtllib)

    def add_block(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        """Add a block to the IDF.

        :param name: A name for the block.
        :param coordinates: A list of (x, y) tuples representing the building outline.
        :param height: The height of the block roof above ground level.
        :param num_stories: The total number of stories including basement stories. Default : 1.
        :param below_ground_stories: The number of stories below ground. Default : 0.
        :param below_ground_storey_height: The height of each basement storey. Default : 2.5.
        :param zoning: The zoning pattern of the block. Default : by_storey
        :param perim_depth: Depth of the perimeter zones if the core/perim zoning pattern is requested. Default : 3.0.

        """
        block = Block(*args, **kwargs)
        block.zoning = kwargs.get("zoning", "by_storey")
        if block.zoning == "by_storey":
            zones = [
                Zone("Block %s Storey %i" % (block.name, storey["storey_no"]), storey)
                for storey in block.stories
            ]
        elif block.zoning == "core/perim":
            zones = []
            try:
                for name, coords in core_perim_zone_coordinates(
                    block.coordinates, block.perim_depth
                )[0].iteritems():
                    block = Block(
                        name=name,
                        coordinates=coords,
                        height=block.height,
                        num_stories=block.num_stories,
                    )
                    zones += [
                        Zone(
                            "Block %s Storey %i" % (block.name, storey["storey_no"]),
                            storey,
                        )
                        for storey in block.stories
                    ]
            except NotImplementedError:
                raise ValueError("Perimeter depth is too great")
        else:
            raise ValueError("%s is not a valid zoning rule" % block.zoning)

        for zone in zones:
            self.add_zone(zone)

    def add_shading_block(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        """Add a shading block to the IDF.

        :param name: A name for the block.
        :param coordinates: A list of (x, y) tuples representing the building outline.
        :param height: The height of the block roof above ground level.
        :param num_stories: The total number of stories including basement stories. Default : 1.
        :param below_ground_stories: The number of stories below ground. Default : 0.
        :param below_ground_storey_height: The height of each basement storey. Default : 2.5.

        """
        block = Block(*args, **kwargs)
        for i, wall in enumerate(block.walls[0], 1):
            if wall.area <= 0:
                continue
            s = self.newidfobject(
                "SHADING:SITE:DETAILED", Name="%s_%s" % (block.name, i)
            )
            try:
                s.setcoords(wall)
            except ZeroDivisionError:
                self.removeidfobject(s)

    def add_zone(self, zone):
        # type: (Zone) -> None
        """Add a zone to the IDF.

        :param zone: A Zone object holding details about the zone.

        """
        try:
            ggr = self.idfobjects["GLOBALGEOMETRYRULES"][
                0
            ]  # type: Optional[Dict[str, Idf_MSequence]]
        except IndexError:
            ggr = None
        # add zone object
        self.newidfobject("ZONE", Name=zone.name)

        for surface_type in zone.__dict__.keys():
            if surface_type == "name":
                continue
            for i, surface_coords in enumerate(zone.__dict__[surface_type], 1):
                if not surface_coords:
                    continue
                name = "{name} {s_type} {num:04d}".format(
                    name=zone.name, s_type=surface_type[:-1].title(), num=i
                )
                s = self.newidfobject(
                    "BUILDINGSURFACE:DETAILED",
                    Name=name,
                    Surface_Type=surface_type[:-1],
                    Zone_Name=zone.name,
                )
                s.setcoords(surface_coords, ggr)



"""Tests for intersecting and matching."""
import itertools

import pytest
from eppy.iddcurrent import iddcurrent
from six import StringIO

from geomeppy.geom.surfaces import minimal_set
from geomeppy.idf import IDF
from geomeppy.geom.intersect_match import (
    get_adjacencies,
    intersect_idf_surfaces,
    match_idf_surfaces,
)
from geomeppy.geom.polygons import intersect, is_hole, Polygon3D
from geomeppy.recipes import translate_coords
from geomeppy.utilities import almostequal


class TestSetCoords:
    def test_set_coords(self, base_idf):
        # type: (IDF) -> None
        idf = base_idf
        ggr = idf.idfobjects["GLOBALGEOMETRYRULES"]
        wall = idf.idfobjects["BUILDINGSURFACE:DETAILED"][0]
        poly1 = Polygon3D([(0, 1, 0), (0, 0, 0), (1, 0, 0), (1, 1, 0)])
        wall.setcoords(poly1, ggr)


class TestSimpleTestPolygons:
    def test_simple_match(self):
        # type: () -> None
        """
        The intersect function should just return the two polygons.
         ___  
        | 1 |
        |_2_| 

        """
        poly1 = Polygon3D([(0, 1, 0), (0, 0, 0), (1, 0, 0), (1, 1, 0)])
        poly2 = Polygon3D([(1, 1, 0), (1, 0, 0), (0, 0, 0), (0, 1, 0)])
        adjacencies = [(poly1, poly2)]
        result = intersect(*adjacencies[0])
        expected = [poly1, poly2]
        assert len(result) == len(expected)
        for poly in expected:
            assert poly in result

    def test_simple_overlap(self):
        # type: () -> None
        """
        The intersect function should return four surfaces.
         __ ___ __ 
        | 1| 1 |  |  
        |__|_2_|_2| 

         __ ___ __ 
        | 1| 3 |  |  
        |__|_4_|_2| 

        """
        poly1 = Polygon3D([(0, 1, 0), (0, 0, 0), (2, 0, 0), (2, 1, 0)])
        poly2 = Polygon3D([(3, 1, 0), (3, 0, 0), (1, 0, 0), (1, 1, 0)])
        adjacencies = [(poly1, poly2)]
        poly1 = Polygon3D([(0, 1, 0), (0, 0, 0), (1, 0, 0), (1, 1, 0)])
        poly2 = Polygon3D([(3, 1, 0), (3, 0, 0), (2, 0, 0), (2, 1, 0)])
        poly3 = Polygon3D([(1, 1, 0), (1, 0, 0), (2, 0, 0), (2, 1, 0)])
        poly4 = Polygon3D([(2, 1, 0), (2, 0, 0), (1, 0, 0), (1, 1, 0)])
        result = intersect(*adjacencies[0])
        expected = [poly1, poly2, poly3, poly4]
        assert len(result) == len(expected)
        for poly in expected:
            assert poly in result

    def test_simple_hole(self):
        # type: () -> None
        """
         _________ 
        | 1 ___   |  
        |  | 2 |  |
        |  |___|  | 
        |_________|
        
         ________ 
        |\ ___   |  
        |1| 2 | 4|
        | |_3_|  | 
        |/_______|
        
        """
        poly1 = Polygon3D([(0, 4, 0), (0, 0, 0), (4, 0, 0), (4, 4, 0)])
        poly2 = Polygon3D([(2, 2, 0), (2, 1, 0), (1, 1, 0), (1, 2, 0)])
        adjacencies = [(poly1, poly2)]

        poly1 = Polygon3D(
            [(0, 4, 0), (0, 0, 0), (1, 1, 0), (1, 2, 0)]
        )  # smaller section
        poly2 = Polygon3D([(1, 2, 0), (1, 1, 0), (2, 1, 0), (2, 2, 0)])  # inverse hole
        poly3 = Polygon3D([(2, 2, 0), (2, 1, 0), (1, 1, 0), (1, 2, 0)])  # hole
        poly4 = Polygon3D(
            [
                (4, 4, 0),
                (0, 4, 0),
                (1, 2, 0),
                (2, 2, 0),
                (2, 1, 0),
                (1, 1, 0),
                (0, 0, 0),
                (4, 0, 0),
            ]
        )  # larger section
        result = intersect(*adjacencies[0])
        expected = [poly1, poly2, poly3, poly4]
        assert len(result) == len(expected)
        for poly in expected:
            assert poly in result

    def test_three_overlapping(self):
        # type: () -> None
        """
         __ ___ __ __
        | 1| 1 | 3| 3|  
        |__|_2_|_2|__| 

         __ ___ __ __
        | 1| 2 | 4| 6|  
        |__|_3_|_5|__| 
        
        """
        poly1 = Polygon3D([(0, 1, 0), (0, 0, 0), (2, 0, 0), (2, 1, 0)])
        poly2 = Polygon3D([(3, 1, 0), (3, 0, 0), (1, 0, 0), (1, 1, 0)])
        poly3 = Polygon3D([(2, 1, 0), (2, 0, 0), (4, 0, 0), (4, 1, 0)])
        adjacencies = [(poly1, poly2), (poly2, poly3)]
        poly1 = Polygon3D([(0, 1, 0), (0, 0, 0), (1, 0, 0), (1, 1, 0)])
        poly2 = Polygon3D([(2, 1, 0), (2, 0, 0), (1, 0, 0), (1, 1, 0)])
        poly3 = Polygon3D([(1, 1, 0), (1, 0, 0), (2, 0, 0), (2, 1, 0)])
        poly4 = Polygon3D([(2, 1, 0), (2, 0, 0), (3, 0, 0), (3, 1, 0)])
        poly5 = Polygon3D([(3, 1, 0), (3, 0, 0), (2, 0, 0), (2, 1, 0)])
        poly6 = Polygon3D([(3, 1, 0), (3, 0, 0), (4, 0, 0), (4, 1, 0)])
        expected = [poly1, poly2, poly3, poly4, poly5, poly6]

        result = intersect(*adjacencies[0])
        result.extend(intersect(*adjacencies[1]))

        assert len(minimal_set(result)) == len(minimal_set(expected))
        for poly in expected:
            assert poly in result

    def test_double_overlap(self):
        # type: () -> None
        """
         __________
        |__1_______| 
        | 1 | 2 |1 |
        |_2_|   |2_|
        |__________|
        
         __________
        |__1________| 
        | 2 | 4 | 5 |
        |_3_|   |_6_|
        |___________|
        
        """
        poly1 = Polygon3D([(0, 2, 0), (0, 0, 0), (3, 0, 0), (3, 2, 0)])
        poly2 = Polygon3D(
            [
                (3, 3, 0),
                (3, 1, 0),
                (2, 1, 0),
                (2, 2, 0),
                (1, 2, 0),
                (1, 1, 0),
                (0, 1, 0),
                (0, 3, 0),
            ]
        )
        adjacencies = [(poly1, poly2)]

        poly1 = Polygon3D([(3, 3, 0), (3, 2, 0), (0, 2, 0), (0, 3, 0)])
        poly2 = Polygon3D([(0, 2, 0), (0, 1, 0), (1, 1, 0), (1, 2, 0)])
        poly3 = Polygon3D([(1, 2, 0), (1, 1, 0), (0, 1, 0), (0, 2, 0)])
        poly4 = Polygon3D(
            [
                (1, 2, 0),
                (1, 1, 0),
                (0, 1, 0),
                (0, 0, 0),
                (3, 0, 0),
                (3, 1, 0),
                (2, 1, 0),
                (2, 2, 0),
            ]
        )
        poly5 = Polygon3D([(2, 2, 0), (2, 1, 0), (3, 1, 0), (3, 2, 0)])
        poly6 = Polygon3D([(3, 2, 0), (3, 1, 0), (2, 1, 0), (2, 2, 0)])

        result = intersect(*adjacencies[0])
        expected = [poly1, poly2, poly3, poly4, poly5, poly6]
        assert len(result) == len(expected)
        assert len(result) == len(expected)
        for poly in expected:
            assert poly in result

    def test_vertically_offset(self):
        # type: () -> None
        """
         ___
        |_1_|
        | 1 |
        |_2_|
        |_2_|
         ___
        |_1_|
        | 2 |
        |_3_|
        |_4_|
         
        """
        poly1 = Polygon3D([(0, 0, 1), (0, 0, 0), (2, 0, 0), (2, 0, 1)])
        poly2 = Polygon3D([(3, 0, 1), (3, 0, 0), (1, 0, 0), (1, 0, 1)])
        adjacencies = [(poly1, poly2)]
        poly1 = Polygon3D([(0, 0, 1), (0, 0, 0), (1, 0, 0), (1, 0, 1)])
        poly2 = Polygon3D([(3, 0, 1), (3, 0, 0), (2, 0, 0), (2, 0, 1)])
        poly3 = Polygon3D([(1, 0, 1), (1, 0, 0), (2, 0, 0), (2, 0, 1)])
        poly4 = Polygon3D([(2, 0, 1), (2, 0, 0), (1, 0, 0), (1, 0, 1)])
        result = intersect(*adjacencies[0])
        expected = [poly1, poly2, poly3, poly4]
        assert len(result) == len(expected)
        for poly in expected:
            assert poly in result


class TestMatchSurfaces:
    def test_match_idf_surfaces(self, base_idf):
        # type: (IDF) -> None
        idf = base_idf
        intersect_idf_surfaces(idf)
        match_idf_surfaces(idf)
        inside_walls = [
            obj
            for obj in idf.idfobjects["BUILDINGSURFACE:DETAILED"]
            if obj.Outside_Boundary_Condition == "surface"
        ]
        assert inside_walls
        for w in inside_walls:
            assert w.Outside_Boundary_Condition_Object != ""

        outside_walls = [
            obj
            for obj in idf.idfobjects["BUILDINGSURFACE:DETAILED"]
            if obj.Outside_Boundary_Condition == "outdoors"
        ]
        assert outside_walls
        for w in outside_walls:
            assert w.Outside_Boundary_Condition_Object == ""

        floors = idf.getsurfaces("floor")
        assert floors
        for f in floors:
            assert f.Outside_Boundary_Condition == "ground"
            assert f.Outside_Boundary_Condition_Object == ""


class TestAdjacencies:
    def test_get_adjacencies(self, base_idf):
        # type: (IDF) -> None
        surfaces = base_idf.getsurfaces()
        adjacencies = get_adjacencies(surfaces)
        assert (u"BuildingSurface:Detailed", u"z1_WALL_0002") in adjacencies
        assert (u"BuildingSurface:Detailed", u"z2_WALL_0004") in adjacencies
        assert len(adjacencies) == 2


def test_intersect():
    # type: () -> None
    poly1 = Polygon3D(
        [(1.0, 2.1, 0.5), (1.0, 2.1, 0.0), (2.0, 2.0, 0.0), (2.0, 2.0, 0.5)]
    )
    poly2 = Polygon3D(
        [(2.5, 1.95, 0.5), (2.5, 1.95, 0.0), (1.5, 2.05, 0.0), (1.5, 2.05, 0.5)]
    )
    intersection = poly1.intersect(poly2)[0]
    assert not is_hole(poly1, intersection)
    assert not is_hole(poly2, intersection)


def test_real_intersect():
    # type: () -> None
    """
    Test that we can make a previously failing test pass by moving to the
    origin first.
    
    """
    poly1 = Polygon3D(
        [
            (526492.65, 185910.65, 6.0),
            (526492.65, 185910.65, 3.0),
            (526489.05, 185916.45, 3.0),
            (526489.05, 185916.45, 6.0),
        ]
    )
    poly2 = Polygon3D(
        [
            (526489.05, 185916.45, 5.0),
            (526489.05, 185916.45, 2.5),
            (526492.65, 185910.65, 2.5),
            (526492.65, 185910.65, 5.0),
        ]
    )
    min_x = min(min(s.xs) for s in [poly1, poly2])
    min_y = min(min(s.ys) for s in [poly1, poly2])
    poly1 = Polygon3D(translate_coords(poly1, [-min_x, -min_y, 0]))
    poly2 = Polygon3D(translate_coords(poly2, [-min_x, -min_y, 0]))
    intersection = Polygon3D(
        translate_coords(poly1.intersect(poly2)[0], [min_x, min_y, 0])
    )
    poly1 = Polygon3D(translate_coords(poly1, [min_x, min_y, 0]))
    poly2 = Polygon3D(translate_coords(poly2, [min_x, min_y, 0]))
    assert not is_hole(poly1, intersection)
    assert not is_hole(poly2, intersection)


def test_is_hole():
    # type: () -> None
    """Test if a surface represents a hole in one of the surfaces.
    """
    # opposite faces (all touching edges)
    poly1 = Polygon3D([(0, 4, 0), (0, 0, 0), (4, 0, 0), (4, 4, 0)])
    poly2 = Polygon3D(reversed([(0, 4, 0), (0, 0, 0), (4, 0, 0), (4, 4, 0)]))
    intersection = poly1.intersect(poly2)[0]
    assert not is_hole(poly1, intersection)
    assert not is_hole(poly2, intersection)

    # poly2 is within poly1 and reversed (no touching edges)
    poly1 = Polygon3D([(0, 4, 0), (0, 0, 0), (4, 0, 0), (4, 4, 0)])
    poly2 = Polygon3D(reversed([(1, 3, 0), (1, 1, 0), (3, 1, 0), (3, 3, 0)]))
    intersection = poly1.intersect(poly2)[0]
    assert is_hole(poly1, intersection)
    assert not is_hole(poly2, intersection)

    # poly2 is within poly1 and reversed (touches at x=0)
    poly1 = Polygon3D([(0, 4, 0), (0, 0, 0), (4, 0, 0), (4, 4, 0)])
    poly2 = Polygon3D(reversed([(0, 3, 0), (0, 1, 0), (3, 1, 0), (3, 3, 0)]))
    intersection = poly1.intersect(poly2)[0]
    assert not is_hole(poly1, intersection)
    assert not is_hole(poly2, intersection)

    # poly2 overlaps poly1
    poly1 = Polygon3D([(1, 4, 0), (1, 0, 0), (5, 0, 0), (5, 4, 0)])
    poly2 = Polygon3D(reversed([(0, 3, 0), (0, 1, 0), (3, 1, 0), (3, 3, 0)]))
    intersection = poly1.intersect(poly2)[0]
    assert not is_hole(poly1, intersection)
    assert not is_hole(poly2, intersection)


class TestIntersectMatchRing:
    def test_intersect_idf_surfaces(self, ring_idf):
        # type: (IDF) -> None
        idf = ring_idf
        starting = len(idf.idfobjects["BUILDINGSURFACE:DETAILED"])
        intersect_idf_surfaces(idf)
        idf.set_default_constructions()
        ending = len(idf.idfobjects["BUILDINGSURFACE:DETAILED"])
        assert starting == 12
        assert ending == 14
        for name in [
            "z1 Roof 0001_1",
            "z1 Roof 0001_2",
            "z1 Roof 0001_3",
            "z2 Floor 0001_1",
        ]:
            obj = idf.getobject("BUILDINGSURFACE:DETAILED", name)
            assert obj


class TestIntersectMatch:
    def test_getsurfaces(self, base_idf):
        # type: (IDF) -> None
        idf = base_idf
        surfaces = idf.getsurfaces()
        assert len(surfaces) == 12

    def test_intersect_idf_surfaces(self, base_idf):
        # type: (IDF) -> None
        idf = base_idf
        starting = len(idf.idfobjects["BUILDINGSURFACE:DETAILED"])
        intersect_idf_surfaces(idf)
        ending = len(idf.idfobjects["BUILDINGSURFACE:DETAILED"])
        assert starting == 12
        assert ending == 14

        for name in [
            "z1_WALL_0002_1",
            "z1_WALL_0002_2",
            "z2_WALL_0004_1",
            "z2_WALL_0004_2",
        ]:
            obj = idf.getobject("BUILDINGSURFACE:DETAILED", name)
            assert obj


@pytest.mark.xfail("sys.version_info.major == 3 and sys.version_info.minor == 5")
def test_real_scale():
    # type: () -> None
    """Test building, intersecting and matching from a real building footprint.
    """
    iddfhandle = StringIO(iddcurrent.iddtxt)
    if IDF.getiddname() == None:
        IDF.setiddname(iddfhandle)

    idf = IDF(StringIO("Version, 8.5;"))
    poly1 = [
        (526492.65, 185910.65),
        (526489.05, 185916.45),
        (526479.15, 185910.3),
        (526482.65, 185904.6),
        (526492.65, 185910.65),
    ]
    poly2 = [
        (526483.3, 185903.15),
        (526483.5, 185903.25),
        (526482.65, 185904.6),
        (526479.15, 185910.3),
        (526489.05, 185916.45),
        (526492.65, 185910.65),
        (526493.4, 185909.4),
        (526500, 185913.95),
        (526500.45, 185914.3),
        (526500, 185914.85),
        (526497.4, 185918.95),
        (526499.45, 185920.2),
        (526494.4, 185928.35),
        (526466.05, 185910.95),
        (526471.1, 185902.75),
        (526473.05, 185903.9),
        (526476.2, 185898.8),
        (526479.95, 185901.1),
        (526483.3, 185903.15),
    ]
    idf.add_block("small", poly1, 6.0, 2)
    idf.add_block("large", poly2, 5.0, 2)
    idf.translate_to_origin()
    idf.intersect_match()
    idf.set_wwr(0.25)
    walls = idf.getsurfaces("wall")
    # look for a wall which should have been split
    assert "Block large Storey 1 Wall 0003" not in [w.Name for w in walls]
    # look for another wall which should have been split
    assert "Block large Storey 1 Wall 0005" not in [w.Name for w in walls]
    # look for a wall which should be an internal wall
    wall = idf.getobject("BUILDINGSURFACE:DETAILED", "Block small Storey 1 Wall 0002_1")
    assert wall.Outside_Boundary_Condition != "outdoors"
    # look for another wall which should be an internal wall
    wall = idf.getobject("BUILDINGSURFACE:DETAILED", "Block large Storey 1 Wall 0003_2")
    assert wall.Outside_Boundary_Condition != "outdoors"

    walls = idf.getsurfaces("wall")
    # look for walls which are being incorrectly duplicated
    for s1, s2 in itertools.combinations(walls, 2):
        assert not almostequal(s1.coords, s2.coords), "Dupes: '{}' and '{}'".format(
            s1.Name, s2.Name
        )



"""
Segment class, representing a line segment.

"""

from typing import Any, Iterator  # noqa

from .vectors import Vector3D
from ..utilities import almostequal

if False:
    from .polygons import Polygon3D  # noqa


class Segment(object):
    """Line segment in 3D."""

    def __init__(self, *vertices):
        # type: (*Vector3D) -> None
        self.vertices = vertices
        self.p1 = vertices[0]
        self.p2 = vertices[1]

    def __repr__(self):
        # type: () -> str
        class_name = type(self).__name__
        return "{}({}, {})".format(class_name, self.p1, self.p2)

    def __neg__(self):
        # type: () -> Segment
        return Segment(self.p2, self.p1)

    def __iter__(self):
        # type: () -> Iterator
        return (i for i in self.vertices)

    def __eq__(self, other):
        # type: (Any) -> bool
        return self.__dict__ == other.__dict__

    def _is_collinear(self, other):
        # type: (Segment) -> bool
        """Test if a segment is collinear with another segment

        :param other: The other segment.
        :return: True if the segments are collinear else False.
        """
        if almostequal(other, self) or almostequal(other, -self):
            return True
        a = self.p1 - other.p1
        b = self.p1 - other.p2
        angle_between = a.cross(b)
        if almostequal(angle_between, Vector3D(0, 0, 0)):
            return True
        a = self.p2 - other.p1
        b = self.p2 - other.p2
        angle_between = a.cross(b)
        if almostequal(angle_between, Vector3D(0, 0, 0)):
            return True
        return False

    def _on_poly_edge(self, poly):
        # type: (Polygon3D) -> bool
        """Test if segment lies on any edge of a polygon

        :param poly: The polygon to test against.
        :returns: True if segment lies on any edge of the polygon, else False.
        """
        for edge in poly.edges:
            if self._is_collinear(edge):
                return True
        return False



import os
from geomeppy.io.obj import export_to_obj


def test_export_to_obj(base_idf):
    export_to_obj(base_idf, "test.obj")
    assert os.path.isfile("test.obj")
    assert os.path.isfile("test.mtl")
    os.remove("test.obj")
    os.remove("test.mtl")



"""Utilities for use in geomeppy."""
from typing import Any  # noqa

from six.moves import zip


def almostequal(first, second, places=7):
    # type: (Any, Any, int) -> bool
    """Tests a range of types for near equality."""
    try:
        # try converting to float first
        first = float(first)
        second = float(second)
        # test floats for near-equality
        if round(abs(second - first), places) != 0:
            return False
        else:
            return True
    except ValueError:
        # handle non-float types
        return str(first) == str(second)
    except TypeError:
        # handle iterables
        return all([almostequal(a, b, places) for a, b in zip(first, second)])



"""Tool for visualising geometry."""
from typing import Optional  # noqa

from eppy.function_helpers import getcoords
from eppy.iddcurrent import iddcurrent
from eppy.modeleditor import IDF
from six import StringIO
from six.moves.tkinter import TclError

try:
    from mpl_toolkits.mplot3d import Axes3D  # noqa
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection
    import matplotlib.pyplot as plt
except (ImportError, RuntimeError):
    # this isn't always needed so we can ignore if it's not present
    pass


def view_idf(fname=None, idf_txt=None, test=False):
    # type: (Optional[str], Optional[str], Optional[bool]) -> None
    """Display an IDF for inspection.

    :param fname: Path to the IDF.
    :param idf_txt: The string representation of an IDF.
    """
    try:
        plt.figure()
    except TclError:
        # this is as expected on the test server
        return
    if fname and idf_txt:
        raise ValueError("Pass either fname or idf_txt, not both.")
    # set the IDD for the version of EnergyPlus
    iddfhandle = StringIO(iddcurrent.iddtxt)
    if IDF.getiddname() is None:
        IDF.setiddname(iddfhandle)

    if fname:
        # import the IDF
        idf = IDF(fname)
    elif idf_txt:
        idf = IDF()
        idf.initreadtxt(idf_txt)
    # create the figure and add the surfaces
    ax = plt.axes(projection="3d")
    collections = _get_collections(idf, opacity=0.5)
    for c in collections:
        ax.add_collection3d(c)

    # calculate and set the axis limits
    limits = _get_limits(idf=idf)
    ax.set_xlim(limits["x"])
    ax.set_ylim(limits["y"])
    ax.set_zlim(limits["z"])

    if not test:
        plt.show()


def view_polygons(polygons):
    """Display a collection of polygons for inspection.

    :param polygons: A dict keyed by colour, containing Polygon3D objects to show in that colour.
    """
    # create the figure and add the surfaces
    plt.figure()
    ax = plt.axes(projection="3d")

    collections = _make_collections(polygons, opacity=0.5)

    for c in collections:
        ax.add_collection3d(c)

    # calculate and set the axis limits
    limits = _get_limits(polygons=polygons)
    ax.set_xlim(limits["x"])
    ax.set_ylim(limits["y"])
    ax.set_zlim(limits["z"])

    plt.show()


def _get_surfaces(idf):
    """Get the surfaces from the IDF.
    """
    surface_types = ["BUILDINGSURFACE:DETAILED", "FENESTRATIONSURFACE:DETAILED"]
    surfaces = []
    for surface_type in surface_types:
        surfaces.extend(idf.idfobjects[surface_type])

    return surfaces


def _get_shading(idf):
    """Get the shading surfaces from the IDF."""
    shading_types = ["SHADING:ZONE:DETAILED", "SHADING:SITE:DETAILED"]
    shading = []
    for shading_type in shading_types:
        shading.extend(idf.idfobjects[shading_type])

    return shading


def _get_collections(idf, opacity=1):
    """Set up 3D collections for each surface type."""
    surfaces = _get_surfaces(idf)
    # set up the collections
    walls = _get_collection("wall", surfaces, opacity, facecolor="lightyellow")
    floors = _get_collection("floor", surfaces, opacity, facecolor="dimgray")
    roofs = _get_collection("roof", surfaces, opacity, facecolor="firebrick")
    windows = _get_collection("window", surfaces, opacity, facecolor="cornflowerblue")

    shading_surfaces = _get_shading(idf)
    shading = Poly3DCollection(
        [getcoords(s) for s in shading_surfaces],
        alpha=opacity,
        facecolor="darkolivegreen",
        edgecolors="black",
    )

    return walls, roofs, floors, windows, shading


def _get_collection(surface_type, surfaces, opacity, facecolor, edgecolors="black"):
    """Make collections from a list of EnergyPlus surfaces."""
    coords = [
        getcoords(s) for s in surfaces if s.Surface_Type.lower() == surface_type.lower()
    ]
    trimmed_coords = [c for c in coords if c]  # dump any empty surfaces
    collection = Poly3DCollection(
        trimmed_coords, alpha=opacity, facecolor=facecolor, edgecolors=edgecolors
    )
    return collection


def _make_collections(polygons, opacity=1):
    """Make collections from a dict of polygons."""
    collection = []
    for color in polygons:
        collection.append(
            Poly3DCollection(
                [p.points_matrix for p in polygons[color]],
                alpha=opacity,
                facecolor=color,
                edgecolors="black",
            )
        )
    return collection


def _get_limits(idf=None, polygons=None):
    """Get limits for the x, y and z axes so the plot is fitted to the axes."""
    if polygons:
        x = [pt[0] for color in polygons for p in polygons[color] for pt in p]
        y = [pt[1] for color in polygons for p in polygons[color] for pt in p]
        z = [pt[2] for color in polygons for p in polygons[color] for pt in p]

    elif idf:
        surfaces = _get_surfaces(idf)

        x = [pt[0] for s in surfaces for pt in getcoords(s)]
        y = [pt[1] for s in surfaces for pt in getcoords(s)]
        z = [pt[2] for s in surfaces for pt in getcoords(s)]

    max_delta = max((max(x) - min(x)), (max(y) - min(y)), (max(z) - min(z)))

    return {
        "x": (min(x), min(x) + max_delta),
        "y": (min(y), min(y) + max_delta),
        "z": (min(z), min(y) + max_delta),
    }


def main(fname=None, polygons=None):
    if fname:
        view_idf(fname)
    elif polygons:
        view_polygons(polygons)


if __name__ == "__main__":
    main()



"""A collection of functions which act on surfaces or lists of surfaces."""
from collections import defaultdict
from itertools import combinations
from typing import Dict, List, Tuple, Union  # noqa
import warnings

from eppy.bunch_subclass import EpBunch  # noqa
from eppy.idf_msequence import Idf_MSequence  # noqa
from numpy import float64  # noqa
from shapely.geometry import Polygon
from shapely.ops import polygonize
from shapely.ops import unary_union

from geomeppy.geom.polygons import Polygon2D
from .polygons import intersect, Polygon3D
from .vectors import Vector2D, Vector3D  # noqa
from ..utilities import almostequal


def set_coords(
    surface,  # type: EpBunch
    coords,
    # type: Union[List[Vector3D], List[Tuple[float, float, float]], Polygon3D]
    ggr,  # type: Union[List, None, Idf_MSequence]
):
    # type: (...) -> None
    """Update the coordinates of a surface.

    :param surface: The surface to modify.
    :param coords: The new coordinates as lists of [x,y,z] lists.
    :param ggr: Global geometry rules.
    """
    coords = list(coords)
    deduped = [c for i, c in enumerate(coords) if c != coords[(i + 1) % len(coords)]]
    poly = Polygon3D(deduped)
    poly = poly.normalize_coords(ggr)
    coords = [i for vertex in poly for i in vertex]
    if len(coords) > 120:
        warnings.warn(
            "To create surfaces with >120 vertices, ensure you have customised your IDD before running EnergyPlus. "
            "https://unmethours.com/question/9343/energy-idf-parsing-error/?answer=9344#post-id-9344"
        )
    # find the vertex fields
    n_vertices_index = surface.objls.index("Number_of_Vertices")
    first_x = n_vertices_index + 1  # X of first coordinate
    surface.obj = surface.obj[:first_x]
    # set the vertex field values
    surface.fieldvalues.extend(coords)


def set_matched_surfaces(surface, matched):
    # type: (EpBunch, EpBunch) -> None
    """Set boundary conditions for two adjoining surfaces.

    :param surface: The first surface.
    :param matched: The second surface.
    """
    if (
        str(surface.key).upper() == "BUILDINGSURFACE:DETAILED"
        and str(matched.key).upper() == "BUILDINGSURFACE:DETAILED"
    ):
        for s in [surface, matched]:
            s.Outside_Boundary_Condition = "surface"
            s.Sun_Exposure = "NoSun"
            s.Wind_Exposure = "NoWind"
        surface.Outside_Boundary_Condition_Object = matched.Name
        matched.Outside_Boundary_Condition_Object = surface.Name
    elif str(surface.key).upper() == "BUILDINGSURFACE:DETAILED" and str(
        matched.key
    ).upper() in ({"SHADING:SITE:DETAILED", "SHADING:ZONE:DETAILED"}):
        surface.Outside_Boundary_Condition = "adiabatic"
        surface.Sun_Exposure = "NoSun"
        surface.Wind_Exposure = "NoWind"
    elif str(matched.key).upper() == "BUILDINGSURFACE:DETAILED" and str(
        surface.key
    ).upper() in ({"SHADING:SITE:DETAILED", "SHADING:ZONE:DETAILED"}):
        matched.Outside_Boundary_Condition = "adiabatic"
        matched.Sun_Exposure = "NoSun"
        matched.Wind_Exposure = "NoWind"


def set_unmatched_surface(surface, vector):
    # type: (EpBunch, Union[Vector2D, Vector3D]) -> None
    """Set boundary conditions for a surface which does not adjoin another one.

    :param surface: The surface.
    :param vector: The surface normal vector.
    """
    if not hasattr(surface, "View_Factor_to_Ground"):
        return
    surface.View_Factor_to_Ground = "autocalculate"
    poly = Polygon3D(surface.coords)
    if min(poly.zs) < 0 or all(z == 0 for z in poly.zs):
        # below ground or ground-adjacent surfaces
        surface.Outside_Boundary_Condition_Object = ""
        surface.Outside_Boundary_Condition = "ground"
        surface.Sun_Exposure = "NoSun"
        surface.Wind_Exposure = "NoWind"
    else:
        surface.Outside_Boundary_Condition = "outdoors"
        surface.Outside_Boundary_Condition_Object = ""
        surface.Wind_Exposure = "WindExposed"
        if almostequal(vector, (0, 0, -1)):
            # downward facing surfaces
            surface.Sun_Exposure = "NoSun"
        else:
            surface.Sun_Exposure = "SunExposed"  # other external surfaces


def getidfplanes(surfaces):
    # type: (Idf_MSequence) -> Dict[float64, Dict[Union[Vector2D, Vector3D], List[EpBunch]]]
    """Fast access data structure for potentially matched surfaces.

    Get a data structure populated with all the surfaces in the IDF, keyed by their distance from the origin, and their
    normal vector.

    :param surfaces: List of all the surfaces.
    :returns: Mapping to look up IDF surfaces.
    """
    round_factor = 8
    planes = {}  # type: Dict[float64, Dict[Union[Vector2D, Vector3D], List[EpBunch]]]
    for s in surfaces:
        poly = Polygon3D(s.coords)
        rounded_distance = round(poly.distance, round_factor)
        rounded_normal_vector = Vector3D(
            *[round(axis, round_factor) for axis in poly.normal_vector]
        )
        planes.setdefault(rounded_distance, {}).setdefault(
            rounded_normal_vector, []
        ).append(s)
    return planes


def get_adjacencies(surfaces):
    # type: (Idf_MSequence) -> defaultdict
    """Create a dictionary mapping surfaces to their adjacent surfaces.

    :param surfaces: A mutable list of surfaces.
    :returns: Mapping of surfaces to adjacent surfaces.
    """
    adjacencies = defaultdict(list)  # type: defaultdict
    # find all adjacent surfaces
    for s1, s2 in combinations(surfaces, 2):
        adjacencies = populate_adjacencies(adjacencies, s1, s2)
    for adjacency, polys in adjacencies.items():
        adjacencies[adjacency] = minimal_set(polys)
    return adjacencies


def minimal_set(polys):
    """Remove overlaps from a set of polygons.

    :param polys: List of polygons.
    :returns: List of polygons with no overlaps.
    """
    normal = polys[0].normal_vector
    as_2d = [p.project_to_2D() for p in polys]
    as_shapely = [Polygon(p) for p in as_2d]
    lines = [p.boundary for p in as_shapely]
    borders = unary_union(lines)
    shapes = [Polygon2D(p.boundary.coords) for p in polygonize(borders)]
    as_3d = [p.project_to_3D(polys[0]) for p in shapes]
    if not almostequal(as_3d[0].normal_vector, normal):
        as_3d = [p.invert_orientation() for p in as_3d]
    return [p for p in as_3d if p.area > 0]


def populate_adjacencies(adjacencies, s1, s2):
    # type: (defaultdict, EpBunch, EpBunch) -> defaultdict
    """Update the adjacencies dict with any intersections between two surfaces.

    :param adjacencies: Dict to contain lists of adjacent surfaces.
    :param s1: Object representing an EnergyPlus surface.
    :param s2: Object representing an EnergyPlus surface.
    :returns: An updated dict of adjacencies.
    """
    poly1 = Polygon3D(s1.coords)
    poly2 = Polygon3D(s2.coords)
    if not almostequal(abs(poly1.distance), abs(poly2.distance), 4):
        return adjacencies
    if not almostequal(poly1.normal_vector, poly2.normal_vector, 4):
        if not almostequal(poly1.normal_vector, -poly2.normal_vector, 4):
            return adjacencies

    intersection = poly1.intersect(poly2)
    if intersection:
        new_surfaces = intersect(poly1, poly2)
        new_s1 = [
            s
            for s in new_surfaces
            if almostequal(s.normal_vector, poly1.normal_vector, 4)
        ]
        new_s2 = [
            s
            for s in new_surfaces
            if almostequal(s.normal_vector, poly2.normal_vector, 4)
        ]
        adjacencies[(s1.key, s1.Name)] += new_s1
        adjacencies[(s2.key, s2.Name)] += new_s2
    return adjacencies



import pytest
from six import StringIO
from geomeppy.geom.core_perim import core_perim_zone_coordinates, get_core, get_perims
from geomeppy.geom.polygons import Polygon2D
from geomeppy import IDF

footprint = [(0, 0), (30, 0), (30, 20), (0, 20)]
expected_footprint = {
    "Core_Zone": [(5.0, 5.0), (5.0, 15.0), (25.0, 15.0), (25.0, 5.0), (5.0, 5.0)],
    "Perimeter_Zone_4": [(5.0, 15.0), (0.0, 20.0), (0.0, 0.0), (5.0, 5.0)],
    "Perimeter_Zone_3": [(25.0, 15.0), (30.0, 20.0), (0.0, 20.0), (5.0, 15.0)],
    "Perimeter_Zone_2": [(25.0, 5.0), (30.0, 0.0), (30.0, 20.0), (25.0, 15.0)],
    "Perimeter_Zone_1": [(5.0, 5.0), (0.0, 0.0), (30.0, 0.0), (25.0, 5.0)],
}


def test_core_perim():
    perim_depth = 5
    assert core_perim_zone_coordinates(footprint, perim_depth)[0] == expected_footprint
    assert get_core(footprint, perim_depth) == Polygon2D(
        expected_footprint["Core_Zone"]
    )
    for idx, zone in enumerate(
        get_perims(footprint, get_core(footprint, perim_depth)), 1
    ):
        assert zone == Polygon2D(expected_footprint["Perimeter_Zone_%i" % idx])


def test_perim_depth():
    idf_file = IDF(StringIO(""))

    with pytest.raises(ValueError) as excinfo:
        idf_file.add_block(
            name="footprint",
            coordinates=footprint,
            height=3,
            zoning="core/perim",
            num_stories=1,
            perim_depth=10,
        )
    assert str(excinfo.value) == "Perimeter depth is too great"



"""Recipes for making changes to EnergyPlus IDF files.

These are generally exposed and methods on the IDF object, e.g. `set_default_constructions(idf)`
can be called on an existing `IDF` object like ``myidf.set_default_constructions()``.

"""
from typing import List, Optional, Tuple, Union  # noqa
import warnings

from eppy.idf_msequence import Idf_MSequence  # noqa
import numpy as np

from .geom.polygons import Polygon3D
from .geom.transformations import Transformation
from .geom.vectors import Vector2D, Vector3D  # noqa

if False:
    from .idf import IDF  # noqa
if False:
    from .patches import EpBunch  # noqa


def set_default_constructions(idf):
    # type: (IDF) -> None
    """Set default constructions for surfaces in the model.

    :param idf: The IDF object.

    """
    constructions = [
        "Project Wall",
        "Project Partition",
        "Project Floor",
        "Project Flat Roof",
        "Project Ceiling",
        "Project Door",
    ]
    for construction in constructions:
        idf.newidfobject(
            "CONSTRUCTION", Name=construction, Outside_Layer="DefaultMaterial"
        )
    idf.newidfobject(
        "MATERIAL",
        Name="DefaultMaterial",
        Roughness="Rough",
        Thickness=0.1,
        Conductivity=0.1,
        Density=1000,
        Specific_Heat=1000,
    )

    idf.newidfobject(
        "CONSTRUCTION", Name="Project External Window", Outside_Layer="DefaultGlazing"
    )
    idf.newidfobject(
        "WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM",
        Name="DefaultGlazing",
        UFactor=2.7,
        Solar_Heat_Gain_Coefficient=0.763,
        Visible_Transmittance=0.8,
    )

    for surface in idf.getsurfaces():
        set_default_construction(surface)
    for subsurface in idf.getsubsurfaces():
        set_default_construction(subsurface)


def set_default_construction(surface):
    # type: (EpBunch) -> None
    """Set default construction for a surface in the model.

    :param surface: A model surface.

    """
    if surface.Surface_Type.lower() == "wall":
        if surface.Outside_Boundary_Condition.lower() == "outdoors":
            surface.Construction_Name = "Project Wall"
        elif surface.Outside_Boundary_Condition.lower() == "ground":
            surface.Construction_Name = "Project Wall"
        else:
            surface.Construction_Name = "Project Partition"
    if surface.Surface_Type.lower() == "floor":
        if surface.Outside_Boundary_Condition.lower() == "ground":
            surface.Construction_Name = "Project Floor"
        else:
            surface.Construction_Name = "Project Floor"
    if surface.Surface_Type.lower() == "roof":
        surface.Construction_Name = "Project Flat Roof"
    if surface.Surface_Type.lower() == "ceiling":
        surface.Construction_Name = "Project Ceiling"
    if surface.Surface_Type == "window":
        surface.Construction_Name = "Project External Window"
    if surface.Surface_Type == "door":
        surface.Construction_Name = "Project Door"


def set_wwr(
    idf, wwr=0.2, construction=None, force=False, wwr_map=None, orientation=None
):
    # type: (IDF, Optional[float], Optional[str], Optional[bool], Optional[dict], Optional[str]) -> None
    """Set the window to wall ratio on all external walls.

    :param idf: The IDF to edit.
    :param wwr: The window to wall ratio.
    :param construction: Name of a window construction.
    :param force: True to remove all subsurfaces before setting the WWR.
    :param wwr_map: Mapping from wall orientation (azimuth) to WWR, e.g. {180: 0.25, 90: 0.2}.
    :param orientation: One of "north", "east", "south", "west". Walls within 45 degrees will be affected.

    """
    try:
        ggr = idf.idfobjects["GLOBALGEOMETRYRULES"][0]  # type: Optional[Idf_MSequence]
    except IndexError:
        ggr = None

    # check orientation
    orientations = {
        "north": 0.0,
        "east": 90.0,
        "south": 180.0,
        "west": 270.0,
        None: None,
    }
    degrees = orientations.get(orientation, None)
    external_walls = filter(
        lambda x: x.Outside_Boundary_Condition.lower() == "outdoors",
        idf.getsurfaces("wall"),
    )
    external_walls = filter(
        lambda x: _has_correct_orientation(x, degrees), external_walls
    )
    subsurfaces = idf.getsubsurfaces()
    base_wwr = wwr
    for wall in external_walls:
        # get any subsurfaces on the wall
        wall_subsurfaces = list(
            filter(lambda x: x.Building_Surface_Name == wall.Name, subsurfaces)
        )
        if not all(_is_window(wss) for wss in wall_subsurfaces) and not force:
            raise ValueError(
                'Not all subsurfaces on wall "{name}" are windows. '
                "Use `force=True` to replace all subsurfaces.".format(name=wall.Name)
            )

        if wall_subsurfaces and not construction:
            constructions = list(
                {wss.Construction_Name for wss in wall_subsurfaces if _is_window(wss)}
            )
            if len(constructions) > 1:
                raise ValueError(
                    'Not all subsurfaces on wall "{name}" have the same construction'.format(
                        name=wall.Name
                    )
                )
            construction = constructions[0]
        # remove all subsurfaces
        for ss in wall_subsurfaces:
            idf.removeidfobject(ss)
        wwr = (wwr_map or {}).get(wall.azimuth) or base_wwr
        if not wwr:
            return
        coords = window_vertices_given_wall(wall, wwr)
        window = idf.newidfobject(
            "FENESTRATIONSURFACE:DETAILED",
            Name="%s window" % wall.Name,
            Surface_Type="Window",
            Construction_Name=construction,
            Building_Surface_Name=wall.Name,
            View_Factor_to_Ground="autocalculate",  # from the surface angle
        )
        window.setcoords(coords, ggr)


def _has_correct_orientation(wall, orientation_degrees):
    # type: (EpBunch, Optional[float]) -> bool
    """Check that the wall has an orientation which requires WWR to be set.

    :param wall: An EpBunch representing a wall.
    :param orientation_degrees: Orientation in degrees.
    :return: True if the wall is within 45 degrees of the orientation passed, or no orientation passed.
             False if the wall is not within 45 of the orientation passed.
    """
    if not orientation_degrees:
        return True
    if abs(wall.azimuth - orientation_degrees) < 45:
        return True
    return False


def _is_window(subsurface):
    if subsurface.key.lower() in {"window", "fenestrationsurface:detailed"}:
        return True


def window_vertices_given_wall(wall, wwr):
    # type: (EpBunch, float) -> Polygon3D
    """Calculate window vertices given wall vertices and glazing ratio.

    :: For each axis:
        1) Translate the axis points so that they are centred around zero
        2) Either:
            a) Multiply the z dimension by the glazing ratio to shrink it vertically
            b) Multiply the x or y dimension by 0.995 to keep inside the surface
        3) Translate the axis points back to their original positions

    :param wall: The wall to add a window on. We expect each wall to have four vertices.
    :param wwr: Window to wall ratio.

    :returns: Window vertices bounding a vertical strip midway up the surface.

    """
    vertices = wall.coords
    average_x = sum([x for x, _y, _z in vertices]) / len(vertices)
    average_y = sum([y for _x, y, _z in vertices]) / len(vertices)
    average_z = sum([z for _x, _y, z in vertices]) / len(vertices)
    # move windows in 0.5% from the edges so they can be drawn in SketchUp
    window_points = [
        [
            ((x - average_x) * 0.999) + average_x,
            ((y - average_y) * 0.999) + average_y,
            ((z - average_z) * wwr) + average_z,
        ]
        for x, y, z in vertices
    ]

    return Polygon3D(window_points)


def translate_to_origin(idf):
    # type: (IDF) -> None
    """Move an IDF close to the origin so that it can be viewed in SketchUp.

    :param idf: The IDF to edit.

    """
    surfaces = idf.getsurfaces()
    subsurfaces = idf.getsubsurfaces()
    shading_surfaces = idf.getshadingsurfaces()

    min_x = min(min(Polygon3D(s.coords).xs) for s in surfaces)
    min_y = min(min(Polygon3D(s.coords).ys) for s in surfaces)

    translate(surfaces, (-min_x, -min_y))
    translate(subsurfaces, (-min_x, -min_y))
    translate(shading_surfaces, (-min_x, -min_y))


def translate(
    surfaces, vector
):  # type: (Idf_MSequence, Union[Tuple[float, float], Vector2D, Vector3D]) -> None
    """Translate all surfaces by a vector.

    :param surfaces: A list of EpBunch objects.
    :param vector: Representation of a vector to translate by.

    """
    vector = Vector3D(*vector)
    for s in surfaces:
        if not s.coords:
            warnings.warn(
                "%s was not affected by this operation since it does not define vertices."
                % s.Name
            )
            continue
        new_coords = translate_coords(s.coords, vector)
        s.setcoords(new_coords)


def translate_coords(coords, vector):
    # type: (Union[List[Tuple[float, float, float]], Polygon3D], Union[List[float], Vector3D]) -> List[Union[Vector2D, Vector3D]]
    """Translate a set of coords by a direction vector.

    :param coords: A list of points.
    :param vector: Representation of a vector to translate by.
    :returns: List of translated vectors.

    """
    return [Vector3D(*v) + vector for v in coords]


def scale(surfaces, factor, axes):
    # type: (Idf_MSequence, float, str) -> None
    """Scale all surfaces by a factor.

    :param surfaces: A list of EpBunch objects.
    :param factor: Factor to scale the surfaces by.
    :param axes: Axes to scale on.

    """
    for s in surfaces:
        if not s.coords:
            warnings.warn(
                "%s was not affected by this operation since it does not define vertices."
                % s.Name
            )
            continue
        new_coords = scale_coords(s.coords, factor, axes)
        s.setcoords(new_coords)


def scale_coords(coords, factor, axes="xy"):
    # type: (Union[List[Tuple[float, float, float]], Polygon3D], float, str) -> Polygon3D
    """Scale a set of coords by a factor.

    :param coords: A list of points.
    :param factor: Factor to scale the surfaces by.
    :param axes: Axes to scale on.
    :returns: A scaled polygon.

    """
    coords = Polygon3D(coords)
    vertices = []
    for coord in coords:
        x = coord[0] * factor if "x" in axes else coord[0]
        y = coord[1] * factor if "y" in axes else coord[1]
        z = coord[2] * factor if "z" in axes else coord[2]
        vertices.append(Vector3D(x, y, z))
    return Polygon3D(vertices)


def rotate(surfaces, angle):
    # type: (Union[List[EpBunch], Idf_MSequence], float) -> None
    """Rotate all surfaces by an angle.

    :param surfaces: A list of EpBunch objects or a mutable sequence.
    :param angle: An angle in degrees.

    """
    radians = np.deg2rad(angle)
    for s in surfaces:
        if not s.coords:
            warnings.warn(
                "%s was not affected by this operation since it does not define vertices."
                % s.Name
            )
            continue
        new_coords = rotate_coords(s.coords, radians)
        s.setcoords(new_coords)


def rotate_coords(coords, radians):
    """Rotate a set of coords by an angle in radians.

    :param coords: A list of points.
    :param radians: The angle to rotate by.
    :returns: List of Vector3D objects.

    """
    poly = Polygon3D(coords)
    rotation = Transformation()._rotation(Vector3D(0, 0, 1), radians)
    coords = rotation * poly
    return coords



import os

from setuptools import setup

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(THIS_DIR, "requirements.in")) as f:
    install_requires = [line for line in f if line and line[0] not in "#-"]

with open(os.path.join(THIS_DIR, "test-requirements.in")) as f:
    test_requires = [line for line in f if line and line[0] not in "#-"]

with open(os.path.join(THIS_DIR, "README.md")) as f:
    long_description = f.read()

setup(
    name="geomeppy",
    packages=["geomeppy", "geomeppy.geom", "geomeppy.io", "tests"],
    version="0.11.4",
    description="Geometry editing for E+ idf files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jamie Bull",
    author_email="jamie.bull@oco-carbon.com",
    url="https://github.com/jamiebull1/geomeppy",
    download_url="https://github.com/jamiebull1/geomeppy/tarball/v0.11.4",
    license="MIT License",
    keywords=["EnergyPlus", "geometry", "building performance simulation"],
    platforms="any",
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Natural Language :: English",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
    ],
    extras_require={
        ':python_version>="3.7"': ["mypy", "black"],
        "testing": test_requires,
    },  # static type checking
    include_package_data=True,
)



"""Intersect and match all surfaces in an IDF.
"""
from itertools import product

from geomeppy.geom.surfaces import (
    get_adjacencies,
    getidfplanes,
    set_coords,
    set_matched_surfaces,
    set_unmatched_surface,
)
from geomeppy.utilities import almostequal

if False:
    from ..idf import IDF  # noqa


def intersect_idf_surfaces(idf):
    # type: (IDF) -> None
    """Intersect all surfaces in an IDF.

    :param idf: The IDF.
    """
    surfaces = idf.getsurfaces() + idf.getshadingsurfaces()
    try:
        ggr = idf.idfobjects["GLOBALGEOMETRYRULES"][0]
    except IndexError:
        ggr = None
    # get all the intersected surfaces
    adjacencies = get_adjacencies(surfaces)
    for surface in adjacencies:
        key, name = surface
        new_surfaces = adjacencies[surface]
        old_obj = idf.getobject(key.upper(), name)
        for i, new_coords in enumerate(new_surfaces, 1):
            new = idf.copyidfobject(old_obj)
            new.Name = "%s_%i" % (name, i)
            set_coords(new, new_coords, ggr)
        idf.removeidfobject(old_obj)


def match_idf_surfaces(idf):
    # type: (IDF) -> None
    """Match all surfaces in an IDF.

    :param idf: The IDF.
    """
    surfaces = idf.getsurfaces() + idf.getshadingsurfaces()
    planes = getidfplanes(surfaces)
    matched = {}
    for distance in planes:
        for vector in planes[distance]:
            surfaces = planes[distance][vector]
            for surface in surfaces:
                set_unmatched_surface(surface, vector)
            matches = planes.get(-distance, {}).get(-vector, [])
            for s, m in product(surfaces, matches):
                if almostequal(s.coords, reversed(m.coords)):
                    matched[sorted_tuple(m, s)] = (m, s)

    for key in matched:
        set_matched_surfaces(*matched[key])


def sorted_tuple(m, s):
    """Used as a key for the matches."""
    return tuple(sorted(((s.key, s.Name), (m.key, m.Name))))



"""Heavy lifting geometry for IDF surfaces."""
from collections import MutableSequence
from itertools import product
from math import atan2, pi
from typing import Any, List, Optional, Tuple, Union  # noqa

from eppy.geometry.surface import area
from eppy.idf_msequence import Idf_MSequence  # noqa
import numpy as np
from shapely import wkt
from shapely.geometry.polygon import Polygon as SPoly
from six.moves import zip

from .clippers import Clipper2D, Clipper3D
from .segments import Segment
from .transformations import align_face, invert_align_face
from .vectors import Vector2D, Vector3D
from ..utilities import almostequal


class Polygon(Clipper2D, MutableSequence):
    """Base class for 2D and 3D polygons."""

    @property
    def n_dims(self):
        pass

    @property
    def vector_class(self):
        pass

    @property
    def normal_vector(self):
        pass

    def __init__(self, vertices):
        # type: (Any) -> None
        super(Polygon, self).__init__()
        self.vertices = [self.vector_class(*v) for v in vertices]
        self.as_2d = Polygon2D

    def __repr__(self):
        # type: () -> str
        class_name = type(self).__name__
        return "{}({!r})".format(class_name, self.vertices)

    def __len__(self):
        # type: () -> int
        return len(self.vertices)

    def __delitem__(self, key):
        del self.vertices[key]

    def __getitem__(self, key):
        # type: (Union[int, slice]) -> Any
        return self.vertices[key]

    def __setitem__(self, key, value):
        self.vertices[key] = value

    def __add__(self, other):  # type: (Polygon) -> Union[None, Polygon]
        if len(self) == len(other) and hasattr(other[0], "__len__"):
            # add together two equal polygons
            vertices = [v1 + v2 for v1, v2 in zip(self, other)]
        elif len(self[0]) == len(other):
            # translate by a vector
            vertices = [v + other for v in self]
        else:
            raise ValueError("Incompatible objects: %s + %s" % (self, other))
        return self.__class__(vertices)

    def __sub__(self, other):
        if len(self) == len(other) and hasattr(other[0], "__len__"):
            # subtract two equal polygons
            vertices = [v1 - v2 for v1, v2 in zip(self, other)]
        elif len(self[0]) == len(other):
            # translate by a vector
            vertices = [v - other for v in self]
        else:
            raise ValueError("Incompatible objects: %s + %s" % (self, other))
        return self.__class__(vertices)

    def insert(self, key, value):
        self.vertices.insert(key, value)

    @property
    def area(self):
        # type: () -> np.float64
        return area(self)

    @property
    def bounding_box(self):
        # type: () -> Polygon
        aligned = align_face(self)
        top_left = Vector3D(min(aligned.xs), max(aligned.ys), max(aligned.zs))
        bottom_left = Vector3D(min(aligned.xs), min(aligned.ys), min(aligned.zs))
        bottom_right = Vector3D(max(aligned.xs), min(aligned.ys), min(aligned.zs))
        top_right = Vector3D(max(aligned.xs), max(aligned.ys), max(aligned.zs))

        bbox = Polygon3D([top_left, bottom_left, bottom_right, top_right])
        return invert_align_face(self, bbox)

    def buffer(self, distance=None, join_style=2):
        # type: (Optional[float], Optional[int]) -> Polygon2D
        """Returns a representation of all points within a given distance of the polygon.

        :param join_style: The styles of joins between offset segments: 1 (round), 2 (mitre), and 3 (bevel).

        """
        s_poly = SPoly(self.vertices)
        core = s_poly.buffer(distance=distance, join_style=join_style)
        return Polygon2D(core.boundary.coords)

    @property
    def centroid(self):
        # type: () -> Vector2D
        """The centroid of a polygon."""
        return self.vector_class(
            sum(self.xs) / len(self), sum(self.ys) / len(self), sum(self.zs) / len(self)
        )

    @property
    def edges(self):
        # type: () -> List[Segment]
        """A list of edges represented as Segment objects."""
        vertices = self.vertices
        edges = [
            Segment(vertices[i], vertices[(i + 1) % len(self)])
            for i in range(len(self))
        ]
        return edges

    def invert_orientation(self):
        # type: () -> Polygon
        """Reverse the order of the vertices.

        This can be used to create a matching surface, e.g. the other side of a wall.

        :returns: A polygon.

        """
        return self.__class__(reversed(self.vertices))

    @property
    def is_convex(self):
        return is_convex_polygon(self.vertices_list)
        return False

    @property
    def points_matrix(self):
        # type: () -> np.ndarray
        """Matrix representing the points in a polygon.

        Format::
            [[x1, x2,... xn]
            [y1, y2,... yn]
            [z1, z2,... zn]  # all 0 for 2D polygon

        """
        points = np.zeros((len(self.vertices), self.n_dims))
        for i, v in enumerate(self.vertices):
            points[i, :] = v.as_array(dims=self.n_dims)
        return points

    @property
    def vertices_list(self):
        # type: () -> List[Tuple[float, float, Optional[float]]]
        """A list of the vertices in the format required by pyclipper.

        :returns: A list of tuples like [(x1, y1), (x2, y2),... (xn, yn)].

        """
        return [pt.as_tuple(dims=self.n_dims) for pt in self.vertices]

    @property
    def xs(self):
        # type: () -> List[float]
        return [pt.x for pt in self.vertices]

    @property
    def ys(self):
        # type: () -> List[float]
        return [pt.y for pt in self.vertices]

    @property
    def zs(self):
        pass


class Polygon2D(Polygon):
    """Two-dimensional polygon."""

    n_dims = 2
    vector_class = Vector2D

    def __eq__(self, other):
        if self.__dict__ == other.__dict__:  # try the simple case first
            return True
        else:  # also cover same shape in different rotation
            if self.difference(other):
                return False
            if almostequal(self.normal_vector, other.normal_vector):
                return True
        return False

    @property
    def normal_vector(self):
        # type: () -> Vector3D
        as_3d = Polygon3D((v.x, v.y, 0) for v in self)
        return as_3d.normal_vector

    def project_to_3D(self, example3d):
        # type: (Polygon3D) -> Polygon3D
        """Project the 2D polygon rotated into 3D space.

        This is used to return a previously rotated 3D polygon back to its original orientation, or to to put polygons
        generated from pyclipper into the desired orientation.

        :param example3D: A 3D polygon in the desired plane.
        :returns: A 3D polygon.

        """
        points = self.points_matrix
        proj_axis = example3d.projection_axis
        a = example3d.distance
        v = example3d.normal_vector
        projected_points = project_to_3D(points, proj_axis, a, v)
        return Polygon3D(projected_points)

    @property
    def zs(self):
        # type: () -> List[float]
        return [0.0] * len(self.vertices)


class Polygon3D(Clipper3D, Polygon):
    """Three-dimensional polygon."""

    n_dims = 3
    vector_class = Vector3D

    def __eq__(self, other):
        # check they're in the same plane
        if not almostequal(self.normal_vector, other.normal_vector):
            return False
        if not almostequal(self.distance, other.distance):
            return False
        # if they are in the same plane, check they completely overlap in 2D
        return self.project_to_2D() == other.project_to_2D()

    @property
    def zs(self):
        # type: () -> List[float]
        return [pt.z for pt in self.vertices]

    @property
    def normal_vector(self):
        """Unit normal vector perpendicular to the polygon in the outward direction.

        We use Newell's Method since the cross-product of two edge vectors is not valid for concave polygons.
        https://www.opengl.org/wiki/Calculating_a_Surface_Normal#Newell.27s_Method

        """
        n = [0.0, 0.0, 0.0]

        for i, v_curr in enumerate(self.vertices):
            v_next = self.vertices[(i + 1) % len(self.vertices)]
            n[0] += (v_curr.y - v_next.y) * (v_curr.z + v_next.z)
            n[1] += (v_curr.z - v_next.z) * (v_curr.x + v_next.x)
            n[2] += (v_curr.x - v_next.x) * (v_curr.y + v_next.y)

        return Vector3D(*n).normalize()

    @property
    def distance(self):
        # type: () -> np.float64
        """Distance from the origin to the polygon.

        Where v[0] * x + v[1] * y + v[2] * z = a is the equation of the plane containing the polygon (and where v
        is the polygon normal vector).

        :returns: The distance from the origin to the polygon.

        """
        v = self.normal_vector
        pt = self.points_matrix[0]  # arbitrary point in the polygon
        d = np.dot(v, pt)
        return d

    @property
    def projection_axis(self):
        # type: () -> int
        """An axis which will not lead to a degenerate surface.

        :returns: The axis index.

        """
        proj_axis = max(range(3), key=lambda i: abs(self.normal_vector[i]))
        return proj_axis

    @property
    def is_horizontal(self):
        # type: () -> bool
        """Check if polygon is in the xy plane.

        :returns: True if the polygon is in the xy plane, else False.

        """
        return bool(np.array(self.zs).std() < 1e-8)

    def is_clockwise(self, viewpoint):
        # type: (Vector3D) -> np.bool_
        """Check if vertices are ordered clockwise

        This function checks the vertices as seen from the viewpoint.

        :param viewpoint: A point from which to view the polygon.
        :returns: True if vertices are ordered clockwise when observed from the given viewpoint.

        """
        arbitrary_pt = self.vertices[0]
        v = arbitrary_pt - viewpoint
        n = self.normal_vector
        sign = np.dot(v, n)
        return sign > 0

    def is_coplanar(self, other):
        # type: (Polygon3D) -> bool
        """Check if polygon is in the same plane as another polygon.

        This includes the same plane but opposite orientation.

        :param other: Another polygon.
        :returns: True if the two polygons are coplanar, else False.

        """
        n1 = self.normal_vector
        n2 = other.normal_vector
        d1 = self.distance
        d2 = other.distance

        if almostequal(n1, n2) and almostequal(d1, d2):
            return True
        elif almostequal(n1, -n2) and almostequal(d1, -d2):
            return True
        else:
            return False

    def outside_point(self, entry_direction="counterclockwise"):
        # type: (str) -> Vector3D
        """Return a point outside the zone to which the surface belongs.

        The point will be outside the zone, respecting the global geometry rules
        for vertex entry direction.

        :param entry_direction: Either "clockwise" or "counterclockwise", as seen from outside the space.
        :returns: A point vector.

        """
        entry_direction = entry_direction.lower()
        if entry_direction == "clockwise":
            inside = self.vertices[0] - self.normal_vector
        elif entry_direction == "counterclockwise":
            inside = self.vertices[0] + self.normal_vector
        else:
            raise ValueError("invalid value for entry_direction '%s'" % entry_direction)
        return inside

    def order_points(self, starting_position):
        # type: (str) -> Polygon3D
        """Reorder the vertices based on a starting position rule.

        :param starting_position: The string that defines vertex starting position in EnergyPlus.
        :returns: The reordered polygon.

        """
        if starting_position == "upperleftcorner":
            bbox_corner = self.bounding_box[0]
        elif starting_position == "lowerleftcorner":
            bbox_corner = self.bounding_box[1]
        elif starting_position == "lowerrightcorner":
            bbox_corner = self.bounding_box[2]
        elif starting_position == "upperrightcorner":
            bbox_corner = self.bounding_box[3]
        else:
            raise ValueError("%s is not a valid starting position" % starting_position)
        start_index = self.index(bbox_corner.closest(self))
        new_vertices = [self[(start_index + i) % len(self)] for i in range(len(self))]

        return Polygon3D(new_vertices)

    def project_to_2D(self):
        # type: () -> Polygon2D
        """Project the 3D polygon into 2D space.

        This is so that we can perform operations on it using pyclipper library.

        Project onto either the xy, yz, or xz plane. (We choose the one that
        avoids degenerate configurations, which is the purpose of proj_axis.)

        :returns: A 2D polygon.

        """
        points = self.points_matrix
        projected_points = project_to_2D(points, self.projection_axis)

        return Polygon2D([pt[:2] for pt in projected_points])

    def normalize_coords(self, ggr):
        """Order points, respecting the global geometry rules

        :param ggr: EnergyPlus GlobalGeometryRules object.
        :returns: The normalized polygon.

        """
        try:
            entry_direction = ggr.Vertex_Entry_Direction
        except AttributeError:
            entry_direction = "counterclockwise"
        outside_point = self.outside_point(entry_direction)
        return normalize_coords(self, outside_point, ggr)

    def from_wkt(self, wkt_poly):
        # type: (str) -> Polygon3D
        """Convert a wkt representation of a polygon to GeomEppy.

        This also accounts for the possible presence of inner rings by linking them to the outer ring.

        :param wkt_poly: A text representation of a polygon in well known text (wkt) format.
        :returns: A polygon.

        """
        poly = wkt.loads(wkt_poly)
        exterior = Polygon3D(poly.exterior.coords)
        if poly.interiors:
            # make the exterior into a geomeppy poly
            for inner_ring in poly.interiors:
                # make the interior into a geomeppy poly
                interior = Polygon3D(inner_ring.coords)
                # find the nearest points on the exterior and interior
                links = list(product(interior, exterior))
                links = sorted(links, key=lambda x: x[0].relative_distance(x[1]))
                on_interior = links[0][0]
                on_exterior = links[0][1]
                # join them up
                exterior = Polygon3D(
                    exterior[exterior.index(on_exterior) :]
                    + exterior[: exterior.index(on_exterior) + 1]
                )
                interior = Polygon3D(
                    interior[interior.index(on_interior) :]
                    + interior[: interior.index(on_interior) + 1]
                )
                exterior = Polygon3D(exterior[:] + interior[:])

        return exterior


def break_polygons(poly, hole):
    # type: (Polygon, Polygon) -> List[Polygon]
    """Break up a surface with a hole in it.

    This produces two surfaces, neither of which have a hole in them.

    :param poly: The surface with a hole in.
    :param hole: The hole.
    :returns: Two Polygon3D objects.

    """
    # take the two closest points on the surface perimeter
    links = list(product(poly, hole))
    links = sorted(
        links, key=lambda x: x[0].relative_distance(x[1])
    )  # fast distance check

    first_on_poly = links[0][0]
    last_on_poly = links[1][0]

    first_on_hole = links[1][1]
    last_on_hole = links[0][1]

    new_poly = section(first_on_poly, last_on_poly, poly[:] + poly[:]) + section(
        first_on_hole, last_on_hole, reversed(hole[:] + hole[:])
    )

    new_poly = Polygon3D(new_poly)
    union = hole.union(new_poly)[0]
    new_poly2 = poly.difference(union)[0]
    if not almostequal(new_poly.normal_vector, poly.normal_vector):
        new_poly = new_poly.invert_orientation()
    if not almostequal(new_poly2.normal_vector, poly.normal_vector):
        new_poly2 = new_poly2.invert_orientation()

    return [new_poly, new_poly2]


def section(first, last, coords):
    section_on_hole = []
    for item in coords:
        if item == first:
            section_on_hole.append(item)
        elif section_on_hole:
            section_on_hole.append(item)
            if item == last:
                break
    return section_on_hole


def project(pt, proj_axis):
    # type: (np.ndarray, int) -> Any
    """Project point pt onto either the xy, yz, or xz plane

    We choose the one that avoids degenerate configurations, which is the
    purpose of proj_axis.
    See http://stackoverflow.com/a/39008641/1706564

    """
    return tuple(c for i, c in enumerate(pt) if i != proj_axis)


def project_inv(
    pt, proj_axis, a, v
):  # type: (np.ndarray, int, np.float64, Vector3D) -> Any
    """Returns the vector w in the surface's plane such that project(w) equals x.

    See http://stackoverflow.com/a/39008641/1706564

    :param pt: A two-dimensional point.
    :param proj_axis: The axis to project into.
    :param a: Distance to the origin for the plane to project into.
    :param v: Normal vector of the plane to project into.
    :returns: The transformed point.

    """
    w = list(pt)
    w[proj_axis:proj_axis] = [0.0]
    c = a
    for i in range(3):
        c -= w[i] * v[i]
    c /= v[proj_axis]
    w[proj_axis] = c
    return tuple(w)


def project_to_2D(vertices, proj_axis):
    # type: (np.ndarray, int) -> List[Tuple[np.float64, np.float64]]
    """Project a 3D polygon into 2D space.

    :param vertices: The three-dimensional vertices of the polygon.
    :param proj_axis: The axis to project into.
    :returns: The transformed vertices.

    """
    points = [project(x, proj_axis) for x in vertices]
    return points


def project_to_3D(
    vertices, proj_axis, a, v
):  # type: (np.ndarray, int, np.float64, Vector3D) -> List[Tuple[np.float64, np.float64, np.float64]]
    """Project a 2D polygon into 3D space.

    :param vertices: The two-dimensional vertices of the polygon.
    :param proj_axis: The axis to project into.
    :param a: Distance to the origin for the plane to project into.
    :param v: Normal vector of the plane to project into.
    :returns: The transformed vertices.

    """
    return [project_inv(pt, proj_axis, a, v) for pt in vertices]


def normalize_coords(
    poly, outside_pt, ggr=None
):  # type: (Polygon3D, Vector3D, Union[List, None, Idf_MSequence]) -> Polygon3D
    """Put coordinates into the correct format for EnergyPlus dependent on Global Geometry Rules (GGR).

    :param poly: Polygon with new coordinates, but not yet checked for compliance with GGR.
    :param outside_pt: An outside point of the new polygon.
    :param ggr: EnergyPlus GlobalGeometryRules object.
    :returns: The normalized polygon.

    """
    # check and set entry direction
    poly = set_entry_direction(poly, outside_pt, ggr)
    # check and set starting position
    poly = set_starting_position(poly, ggr)

    return poly


def set_entry_direction(poly, outside_pt, ggr=None):
    """Check and set entry direction for a polygon.

    :param poly: A polygon.
    :param outside_pt: A point beyond the outside face of the polygon.
    :param ggr: EnergyPlus global geometry rules
    :return: A polygon with the vertices correctly oriented.

    """
    if not ggr:
        entry_direction = "counterclockwise"  # EnergyPlus default
    else:
        entry_direction = ggr.Vertex_Entry_Direction.lower()
    if entry_direction == "counterclockwise":
        if poly.is_clockwise(outside_pt):
            poly = poly.invert_orientation()
    elif entry_direction == "clockwise":
        if not poly.is_clockwise(outside_pt):
            poly = poly.invert_orientation()
    return poly


def set_starting_position(poly, ggr=None):
    """Check and set starting position."""
    if not ggr:
        starting_position = "upperleftcorner"  # EnergyPlus default
    else:
        starting_position = ggr.Starting_Vertex_Position.lower()
    poly = poly.order_points(starting_position)

    return poly


def intersect(poly1, poly2):
    # type: (Polygon, Polygon) -> List[Polygon]
    """Calculate the polygons to represent the intersection of two polygons.

    :param poly1: The first polygon.
    :param poly2: The second polygon.
    :returns: A list of unique polygons.

    """
    polys = []  # type: List[Polygon]
    polys.extend(poly1.intersect(poly2))
    polys.extend(poly2.intersect(poly1))
    if is_hole(poly1, poly2):
        polys.extend(break_polygons(poly1, poly2))
    elif is_hole(poly2, poly1):
        polys.extend(break_polygons(poly2, poly1))
    else:
        polys.extend(poly1.difference(poly2))
        polys.extend(poly2.difference(poly1))
    return polys


def is_hole(surface, possible_hole):
    # type: (Polygon, Polygon) -> bool
    """Identify if an intersection is a hole in the surface.

    Check the intersection touches an edge of the surface. If it doesn't then it represents a hole, and this needs
    further processing into valid EnergyPlus surfaces.

    :param surface: The first surface.
    :param possible_hole: The intersection into the surface.
    :returns: True if the possible hole is a hole in the surface.

    """
    if surface.area < possible_hole.area:
        return False
    collinear_edges = (
        edges[0]._is_collinear(edges[1])
        for edges in product(surface.edges, possible_hole.edges)
    )
    return not any(collinear_edges)


def bounding_box(polygons):
    """The bounding box which encompasses all of the polygons in the x,y plane.

    :param polygons: A list of polygons.
    :return: A 2D polygon.

    """
    top_left = (
        min(min(c[0] for c in f.coords) for f in polygons),
        max(max(c[1] for c in f.coords) for f in polygons),
    )
    bottom_left = (
        min(min(c[0] for c in f.coords) for f in polygons),
        min(min(c[1] for c in f.coords) for f in polygons),
    )
    bottom_right = (
        max(max(c[0] for c in f.coords) for f in polygons),
        min(min(c[1] for c in f.coords) for f in polygons),
    )
    top_right = (
        max(max(c[0] for c in f.coords) for f in polygons),
        max(max(c[1] for c in f.coords) for f in polygons),
    )
    return Polygon2D([top_left, bottom_left, bottom_right, top_right])


def is_convex_polygon(polygon):  # noqa
    """Return True if the polynomial defined by the sequence of 2D
    points is 'strictly convex': points are valid, side lengths non-
    zero, interior angles are strictly between zero and a straight
    angle, and the polygon does not intersect itself.

    See: https://stackoverflow.com/a/45372025/1706564

    :: NOTES:
            1.  Algorithm: the signed changes of the direction angles
                from one side to the next side must be all positive or
                all negative, and their sum must equal plus-or-minus
                one full turn (2 pi radians). Also check for too few,
                invalid, or repeated points.
            2.  No check is explicitly done for zero    internal angles
                (180 degree direction-change angle) as this is covered
                in other ways, including the `n < 3` check.

    """
    two_pi = 2 * pi
    try:  # needed for any bad points or direction changes
        # Check for too few points
        if len(polygon) < 3:
            return False
        # Get starting information
        old_x, old_y = polygon[-2]
        new_x, new_y = polygon[-1]
        new_direction = atan2(new_y - old_y, new_x - old_x)
        angle_sum = 0.0
        # Check each point (the side ending there, its angle) and accum. angles
        for ndx, newpoint in enumerate(polygon):
            # Update point coordinates and side directions, check side length
            old_x, old_y, old_direction = new_x, new_y, new_direction
            new_x, new_y = newpoint
            new_direction = atan2(new_y - old_y, new_x - old_x)
            if old_x == new_x and old_y == new_y:
                return False  # repeated consecutive points
            # Calculate & check the normalized direction-change angle
            angle = new_direction - old_direction
            if angle <= -pi:
                angle += two_pi  # make it in half-open interval (-Pi, Pi]
            elif angle > pi:
                angle -= two_pi
            if ndx == 0:  # if first time through loop, initialize orientation
                if angle == 0.0:
                    return False
                orientation = 1.0 if angle > 0.0 else -1.0
            else:  # if other time through loop, check orientation is stable
                if orientation * angle <= 0.0:  # not both pos. or both neg.
                    return False
            # Accumulate the direction-change angle
            angle_sum += angle
        # Check that the total number of full turns is plus-or-minus 1
        return abs(round(angle_sum / two_pi)) == 1
    except (ArithmeticError, TypeError, ValueError):
        return False  # any exception means not a proper convex polygon
